from darts.tests.base_test_class import DartsBaseTestClass
from darts.utils.model_selection import MODEL_AWARE, SIMPLE, train_test_split
from darts.utils.timeseries_generation import constant_timeseries


def make_dataset(rows, cols):
    return [constant_timeseries(value=i, length=cols) for i in range(rows)]


def verify_shape(dataset, rows, cols):
    return len(dataset) == rows and all(len(row) == cols for row in dataset)


class ClassTrainTestSplitTestCase(DartsBaseTestClass):
    def test_parameters_for_axis_0(self):
        train_test_split(make_dataset(2, 10), axis=0, test_size=1)

        # expecting no exception
        self.assertTrue(True)

    def test_parameters_for_axis_1_no_n(self):
        with self.assertRaisesRegex(
            AttributeError,
            "You need to provide non-zero `horizon` and `input_size` parameters when axis=1",
        ):
            train_test_split(
                make_dataset(1, 10), axis=1, horizon=1, vertical_split_type=MODEL_AWARE
            )

    def test_parameters_for_axis_1_no_horizon(self):
        with self.assertRaisesRegex(
            AttributeError,
            "You need to provide non-zero `horizon` and `input_size` parameters when axis=1",
        ):
            train_test_split(
                make_dataset(1, 10),
                axis=1,
                input_size=1,
                vertical_split_type=MODEL_AWARE,
            )

    def test_empty_dataset(self):
        with self.assertRaises(AttributeError):
            train_test_split([])

    def test_horiz_number_of_samples_too_small(self):
        with self.assertRaisesRegex(
            AttributeError, "Not enough data to create training and test sets"
        ):
            train_set, test_set = train_test_split(
                make_dataset(1, 10),
                axis=1,
                input_size=4,
                horizon=7,
                test_size=1,
                vertical_split_type=MODEL_AWARE,
            )

    def test_sunny_day_horiz_split(self):
        train_set, test_set = train_test_split(make_dataset(8, 10))

        self.assertTrue(
            verify_shape(train_set, 6, 10) and verify_shape(test_set, 2, 10),
            "Wrong shapes: training set shape: ({}, {}); test set shape ({}, {})".format(
                len(train_set), len(train_set[0]), len(test_set), len(test_set[0])
            ),
        )

    def test_sunny_day_horiz_split_absolute(self):
        train_set, test_set = train_test_split(make_dataset(8, 10), test_size=2)

        self.assertTrue(
            verify_shape(train_set, 6, 10) and verify_shape(test_set, 2, 10),
            "Wrong shapes: training set shape: ({}, {}); test set shape ({}, {})".format(
                len(train_set), len(train_set[0]), len(test_set), len(test_set[0])
            ),
        )

    def test_horiz_split_overindexing_train_set(self):
        train_set, test_set = train_test_split(make_dataset(8, 10), lazy=True)

        with self.assertRaisesRegex(
            IndexError, "Exceeded the size of the training sequence."
        ):
            train_set[6]

    def test_horiz_split_last_index_train_set(self):
        train_set, test_set = train_test_split(make_dataset(8, 10), lazy=True)
        # no IndexError is thrown
        train_set[5]

    def test_horiz_split_overindexing_test_set(self):
        train_set, test_set = train_test_split(make_dataset(8, 10), lazy=True)

        with self.assertRaisesRegex(
            IndexError, "Exceeded the size of the test sequence."
        ):
            test_set[2]

    def test_horiz_split_last_index_test_set(self):
        train_set, test_set = train_test_split(make_dataset(8, 10))
        # no IndexError is thrown

    # test 6
    def test_sunny_day_vertical_split(self):
        train_set, test_set = train_test_split(
            make_dataset(2, 250),
            axis=1,
            input_size=70,
            horizon=50,
            vertical_split_type=MODEL_AWARE,
        )

        self.assertTrue(
            verify_shape(train_set, 2, 151) and verify_shape(test_set, 2, 169),
            "Wrong shapes: training set shape: ({}, {}); test set shape ({}, {})".format(
                len(train_set), len(train_set[0]), len(test_set), len(test_set[0])
            ),
        )

    # test 7
    def test_test_split_absolute_number_horiz(self):
        train_set, test_set = train_test_split(make_dataset(4, 10), axis=0, test_size=2)

        self.assertTrue(
            verify_shape(train_set, 2, 10) and verify_shape(test_set, 2, 10)
        )

    # test 8
    def test_test_split_absolute_number_vertical(self):
        train_set, test_set = train_test_split(
            make_dataset(4, 10),
            axis=1,
            test_size=2,
            input_size=1,
            horizon=2,
            vertical_split_type=MODEL_AWARE,
        )

        self.assertTrue(
            verify_shape(train_set, 4, 7) and verify_shape(test_set, 4, 4),
            "Wrong shapes: training set shape: ({}, {}); test set shape ({}, {})".format(
                len(train_set), len(train_set[0]), len(test_set), len(test_set[0])
            ),
        )

    def test_negative_test_start_index(self):
        with self.assertRaisesRegex(
            AttributeError, "Not enough data to create training and test sets"
        ):
            train_set, test_set = train_test_split(
                make_dataset(1, 10),
                axis=1,
                input_size=2,
                horizon=9,
                test_size=1,
                vertical_split_type=MODEL_AWARE,
            )

    def test_horiz_split_horizon_equal_to_ts_length(self):
        with self.assertRaisesRegex(
            AttributeError, "Not enough data to create training and test sets"
        ):
            train_set, test_set = train_test_split(
                make_dataset(1, 10),
                axis=1,
                input_size=2,
                horizon=10,
                test_size=1,
                vertical_split_type=MODEL_AWARE,
            )

    def test_single_timeseries_no_horizon_no_n(self):
        with self.assertRaises(AttributeError):
            # even if the default axis is 0, but since it is a single timeseries, default axis is 1
            train_test_split(
                constant_timeseries(value=123, length=10),
                test_size=2,
                vertical_split_type=MODEL_AWARE,
            )

    def test_single_timeseries_sunny_day(self):
        train_set, test_set = train_test_split(
            constant_timeseries(value=123, length=10),
            test_size=2,
            input_size=1,
            horizon=2,
            vertical_split_type=MODEL_AWARE,
        )

        self.assertTrue(
            len(train_set) == 7 and len(test_set) == 4,
            "Wrong shapes: training set shape: {}; test set shape {}".format(
                len(train_set), len(test_set)
            ),
        )

    def test_multi_timeseries_variable_ts_length_sunny_day(self):
        data = [
            constant_timeseries(value=123, length=10),
            constant_timeseries(value=123, length=100),
            constant_timeseries(value=123, length=1000),
        ]
        train_set, test_set = train_test_split(
            data,
            axis=1,
            test_size=2,
            input_size=1,
            horizon=2,
            vertical_split_type=MODEL_AWARE,
        )
        train_lengths = [len(ts) for ts in train_set]
        test_lengths = [len(ts) for ts in test_set]

        self.assertTrue(
            train_lengths == [7, 97, 997] and test_lengths == [4, 4, 4],
            "Wrong shapes: training set shape: {}; test set shape {}".format(
                train_lengths, test_lengths
            ),
        )

    def test_multi_timeseries_variable_ts_length_one_ts_too_small(self):
        data = [
            constant_timeseries(value=123, length=21),
            constant_timeseries(value=123, length=100),
            constant_timeseries(value=123, length=1000),
        ]

        with self.assertRaisesRegex(
            AttributeError, "Not enough data to create training and test sets"
        ):
            train_set, test_set = train_test_split(
                data,
                axis=1,
                test_size=2,
                input_size=1,
                horizon=18,
                vertical_split_type=MODEL_AWARE,
            )

    def test_simple_vertical_split_sunny_day(self):
        train_set, test_set = train_test_split(
            make_dataset(4, 10), axis=1, vertical_split_type=SIMPLE, test_size=0.2
        )

        self.assertTrue(
            verify_shape(train_set, 4, 8) and verify_shape(test_set, 4, 2),
            "Wrong shapes: training set shape: ({}, {}); test set shape ({}, {})".format(
                len(train_set), len(train_set[0]), len(test_set), len(test_set[0])
            ),
        )

    def test_simple_vertical_split_sunny_day_absolute_split(self):
        train_set, test_set = train_test_split(
            make_dataset(4, 10), axis=1, vertical_split_type=SIMPLE, test_size=2
        )

        self.assertTrue(
            verify_shape(train_set, 4, 8) and verify_shape(test_set, 4, 2),
            "Wrong shapes: training set shape: ({}, {}); test set shape ({}, {})".format(
                len(train_set), len(train_set[0]), len(test_set), len(test_set[0])
            ),
        )

    def test_simple_vertical_split_exception_on_bad_param(self):
        # bad value for vertical_split_type
        with self.assertRaises(AttributeError):
            train_set, test_set = train_test_split(
                make_dataset(4, 10),
                axis=1,
                vertical_split_type="WRONG_VALUE",
                test_size=2,
            )

    def test_simple_vertical_split_test_size_too_large(self):

        with self.assertRaisesRegex(
            AttributeError, "`test_size` is bigger then timeseries length"
        ):
            train_set, test_set = train_test_split(
                make_dataset(4, 10), axis=1, vertical_split_type=SIMPLE, test_size=11
            )

    def test_model_aware_vertical_split_empty_training_set(self):

        with self.assertRaisesRegex(
            AttributeError, "Not enough data to create training and test sets"
        ):
            train_set, test_set = train_test_split(
                make_dataset(4, 10),
                axis=1,
                vertical_split_type=MODEL_AWARE,
                test_size=5,
                horizon=3,
                input_size=2,
            )



import numpy as np

from darts.logging import get_logger
from darts.models import LinearRegressionModel, NaiveSeasonal
from darts.tests.base_test_class import DartsBaseTestClass
from darts.tests.models.forecasting.test_regression_models import dummy_timeseries
from darts.utils.timeseries_generation import constant_timeseries as ct
from darts.utils.timeseries_generation import linear_timeseries as lt

logger = get_logger(__name__)


class TestResidualsTestCase(DartsBaseTestClass):

    np.random.seed(42)

    def test_forecasting_residuals_nocov_output(self):
        model = NaiveSeasonal(K=1)

        # test zero residuals
        constant_ts = ct(length=20)
        residuals = model.residuals(constant_ts)
        np.testing.assert_almost_equal(
            residuals.univariate_values(), np.zeros(len(residuals))
        )

        # test constant, positive residuals
        linear_ts = lt(length=20)
        residuals = model.residuals(linear_ts)
        np.testing.assert_almost_equal(
            np.diff(residuals.univariate_values()), np.zeros(len(residuals) - 1)
        )
        np.testing.assert_array_less(
            np.zeros(len(residuals)), residuals.univariate_values()
        )

    def test_forecasting_residuals_inputs(self):
        # test input types past and/or future covariates

        # dummy covariates and target TimeSeries instances

        target_series, past_covariates, future_covariates = dummy_timeseries(
            length=10,
            n_series=1,
            comps_target=1,
            comps_pcov=1,
            comps_fcov=1,
        )  # outputs Sequences[TimeSeries] and not TimeSeries

        model = LinearRegressionModel(
            lags=4, lags_past_covariates=4, lags_future_covariates=(4, 1)
        )
        model.fit(
            series=target_series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
        )

    def test_forecasting_residuals_cov_output(self):
        # if covariates are constant and the target is constant/linear,
        # residuals should be zero (for a LinearRegression model)

        target_series_1 = ct(value=0.5, length=10)
        target_series_2 = lt(length=10)
        past_covariates = ct(value=0.2, length=10)
        future_covariates = ct(value=0.1, length=10)

        model_1 = LinearRegressionModel(
            lags=1, lags_past_covariates=1, lags_future_covariates=(1, 1)
        )
        model_2 = LinearRegressionModel(
            lags=1, lags_past_covariates=1, lags_future_covariates=(1, 1)
        )
        model_1.fit(
            target_series_1,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
        )
        residuals_1 = model_1.residuals(
            target_series_1,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
        )

        model_2.fit(
            target_series_2,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
        )
        residuals_2 = model_2.residuals(
            target_series_2,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
        )

        # residuals zero
        np.testing.assert_almost_equal(
            residuals_1.univariate_values(), np.zeros(len(residuals_1))
        )

        np.testing.assert_almost_equal(
            residuals_2.univariate_values(), np.zeros(len(residuals_2))
        )

        # if model is trained with covariates, should raise error when covariates are missing in residuals()
        with self.assertRaises(ValueError):
            model_1.residuals(target_series_1)

        with self.assertRaises(ValueError):
            model_1.residuals(target_series_1, past_covariates=past_covariates)

        with self.assertRaises(ValueError):
            model_1.residuals(target_series_1, future_covariates=future_covariates)



import os
import shutil
import tempfile
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest

from darts import TimeSeries
from darts.logging import get_logger
from darts.metrics import mape
from darts.tests.base_test_class import DartsBaseTestClass

logger = get_logger(__name__)

try:
    import torch
    from pytorch_lightning.tuner.lr_finder import _LRFinder
    from torchmetrics import (
        MeanAbsoluteError,
        MeanAbsolutePercentageError,
        MetricCollection,
    )

    from darts.models.forecasting.rnn_model import RNNModel
    from darts.utils.likelihood_models import GaussianLikelihood

    TORCH_AVAILABLE = True
except ImportError:
    logger.warning("Torch not available. RNN tests will be skipped.")
    TORCH_AVAILABLE = False

if TORCH_AVAILABLE:

    class TestTorchForecastingModel(DartsBaseTestClass):
        def setUp(self):
            self.temp_work_dir = tempfile.mkdtemp(prefix="darts")

            times = pd.date_range("20130101", "20130410")
            pd_series = pd.Series(range(100), index=times)
            self.series = TimeSeries.from_series(pd_series)

            df = pd.DataFrame({"var1": range(100), "var2": range(100)}, index=times)
            self.multivariate_series = TimeSeries.from_dataframe(df)

        def tearDown(self):
            shutil.rmtree(self.temp_work_dir)

        def test_save_model_parameters(self):
            # check if re-created model has same params as original
            model = RNNModel(12, "RNN", 10, 10)
            self.assertTrue(model._model_params, model.untrained_model()._model_params)

        @patch(
            "darts.models.forecasting.torch_forecasting_model.TorchForecastingModel.save"
        )
        def test_suppress_automatic_save(self, patch_save_model):
            model_name = "test_model"
            model1 = RNNModel(
                12,
                "RNN",
                10,
                10,
                model_name=model_name,
                work_dir=self.temp_work_dir,
                save_checkpoints=False,
            )
            model2 = RNNModel(
                12,
                "RNN",
                10,
                10,
                model_name=model_name,
                work_dir=self.temp_work_dir,
                force_reset=True,
                save_checkpoints=False,
            )

            model1.fit(self.series, epochs=1)
            model2.fit(self.series, epochs=1)

            model1.predict(n=1)
            model2.predict(n=2)

            patch_save_model.assert_not_called()

            model1.save(path=os.path.join(self.temp_work_dir, model_name))
            patch_save_model.assert_called()

        def test_manual_save_and_load(self):
            """validate manual save with automatic save files by comparing output between the two"""

            model_dir = os.path.join(self.temp_work_dir)
            manual_name = "test_save_manual"
            auto_name = "test_save_automatic"
            model_manual_save = RNNModel(
                12,
                "RNN",
                10,
                10,
                model_name=manual_name,
                work_dir=self.temp_work_dir,
                save_checkpoints=False,
                random_state=42,
            )
            model_auto_save = RNNModel(
                12,
                "RNN",
                10,
                10,
                model_name=auto_name,
                work_dir=self.temp_work_dir,
                save_checkpoints=True,
                random_state=42,
            )

            # save model without training
            no_training_ckpt = "no_training.pth.tar"
            no_training_ckpt_path = os.path.join(model_dir, no_training_ckpt)
            model_manual_save.save(no_training_ckpt_path)
            # check that model object file was created
            self.assertTrue(os.path.exists(no_training_ckpt_path))
            # check that the PyTorch Ligthning ckpt does not exist
            self.assertFalse(os.path.exists(no_training_ckpt_path + ".ckpt"))
            # informative exception about `fit()` not called
            with self.assertRaises(
                ValueError,
                msg="The model must be fit before calling predict(). "
                "For global models, if predict() is called without specifying a series, "
                "the model must have been fit on a single training series.",
            ):
                no_train_model = RNNModel.load(no_training_ckpt_path)
                no_train_model.predict(n=4)

            model_manual_save.fit(self.series, epochs=1)
            model_auto_save.fit(self.series, epochs=1)

            # check that file was not created with manual save
            self.assertFalse(
                os.path.exists(os.path.join(model_dir, manual_name, "checkpoints"))
            )
            # check that file was created with automatic save
            self.assertTrue(
                os.path.exists(os.path.join(model_dir, auto_name, "checkpoints"))
            )

            # create manually saved model checkpoints folder
            checkpoint_path_manual = os.path.join(model_dir, manual_name)
            os.mkdir(checkpoint_path_manual)

            checkpoint_file_name = "checkpoint_0.pth.tar"
            model_path_manual = os.path.join(
                checkpoint_path_manual, checkpoint_file_name
            )
            checkpoint_file_name_cpkt = "checkpoint_0.pth.tar.ckpt"
            model_path_manual_ckpt = os.path.join(
                checkpoint_path_manual, checkpoint_file_name_cpkt
            )

            # save manually saved model
            model_manual_save.save(model_path_manual)
            self.assertTrue(os.path.exists(model_path_manual))

            # check that the PTL checkpoint path is also there
            self.assertTrue(os.path.exists(model_path_manual_ckpt))

            # load manual save model and compare with automatic model results
            model_manual_save = RNNModel.load(model_path_manual, map_location="cpu")
            model_manual_save.to_cpu()
            self.assertEqual(
                model_manual_save.predict(n=4), model_auto_save.predict(n=4)
            )

            # load automatically saved model with manual load() and load_from_checkpoint()
            model_auto_save1 = RNNModel.load_from_checkpoint(
                model_name=auto_name,
                work_dir=self.temp_work_dir,
                best=False,
                map_location="cpu",
            )
            model_auto_save1.to_cpu()
            # compare loaded checkpoint with manual save
            self.assertEqual(
                model_manual_save.predict(n=4), model_auto_save1.predict(n=4)
            )

            # save() model directly after load_from_checkpoint()
            checkpoint_file_name_2 = "checkpoint_1.pth.tar"
            checkpoint_file_name_cpkt_2 = checkpoint_file_name_2 + ".ckpt"

            model_path_manual_2 = os.path.join(
                checkpoint_path_manual, checkpoint_file_name_2
            )
            model_path_manual_ckpt_2 = os.path.join(
                checkpoint_path_manual, checkpoint_file_name_cpkt_2
            )
            model_auto_save2 = RNNModel.load_from_checkpoint(
                model_name=auto_name,
                work_dir=self.temp_work_dir,
                best=False,
                map_location="cpu",
            )
            # save model directly after loading, model has no trainer
            model_auto_save2.save(model_path_manual_2)

            # assert original .ckpt checkpoint was correctly copied
            self.assertTrue(os.path.exists(model_path_manual_ckpt_2))

            model_chained_load_save = RNNModel.load(
                model_path_manual_2, map_location="cpu"
            )

            # compare chained load_from_checkpoint() save() with manual save
            self.assertEqual(
                model_chained_load_save.predict(n=4), model_manual_save.predict(n=4)
            )

        def test_create_instance_new_model_no_name_set(self):
            RNNModel(12, "RNN", 10, 10, work_dir=self.temp_work_dir)
            # no exception is raised
            RNNModel(12, "RNN", 10, 10, work_dir=self.temp_work_dir)
            # no exception is raised

        def test_create_instance_existing_model_with_name_no_fit(self):
            model_name = "test_model"
            RNNModel(
                12, "RNN", 10, 10, work_dir=self.temp_work_dir, model_name=model_name
            )
            # no exception is raised

            RNNModel(
                12, "RNN", 10, 10, work_dir=self.temp_work_dir, model_name=model_name
            )
            # no exception is raised

        @patch(
            "darts.models.forecasting.torch_forecasting_model.TorchForecastingModel.reset_model"
        )
        def test_create_instance_existing_model_with_name_force(
            self, patch_reset_model
        ):
            model_name = "test_model"
            RNNModel(
                12, "RNN", 10, 10, work_dir=self.temp_work_dir, model_name=model_name
            )
            # no exception is raised
            # since no fit, there is no data stored for the model, hence `force_reset` does noting

            RNNModel(
                12,
                "RNN",
                10,
                10,
                work_dir=self.temp_work_dir,
                model_name=model_name,
                force_reset=True,
            )
            patch_reset_model.assert_not_called()

        @patch(
            "darts.models.forecasting.torch_forecasting_model.TorchForecastingModel.reset_model"
        )
        def test_create_instance_existing_model_with_name_force_fit_with_reset(
            self, patch_reset_model
        ):
            model_name = "test_model"
            model1 = RNNModel(
                12,
                "RNN",
                10,
                10,
                work_dir=self.temp_work_dir,
                model_name=model_name,
                save_checkpoints=True,
            )
            # no exception is raised

            model1.fit(self.series, epochs=1)

            RNNModel(
                12,
                "RNN",
                10,
                10,
                work_dir=self.temp_work_dir,
                model_name=model_name,
                save_checkpoints=True,
                force_reset=True,
            )
            patch_reset_model.assert_called_once()

        # TODO for PTL: currently we (have to (?)) create a mew PTL trainer object every time fit() is called which
        #  resets some of the model's attributes such as epoch and step counts. We have check whether there is another
        #  way of doing this.

        # n_epochs=20, fit|epochs=None, epochs_trained=0 - train for 20 epochs
        def test_train_from_0_n_epochs_20_no_fit_epochs(self):
            model1 = RNNModel(
                12, "RNN", 10, 10, n_epochs=20, work_dir=self.temp_work_dir
            )

            model1.fit(self.series)

            self.assertEqual(20, model1.epochs_trained)

        # n_epochs = 20, fit|epochs=None, epochs_trained=20 - train for another 20 epochs
        def test_train_from_20_n_epochs_40_no_fit_epochs(self):
            model1 = RNNModel(
                12, "RNN", 10, 10, n_epochs=20, work_dir=self.temp_work_dir
            )

            model1.fit(self.series)
            self.assertEqual(20, model1.epochs_trained)

            model1.fit(self.series)
            self.assertEqual(20, model1.epochs_trained)

        # n_epochs = 20, fit|epochs=None, epochs_trained=10 - train for another 20 epochs
        def test_train_from_10_n_epochs_20_no_fit_epochs(self):
            model1 = RNNModel(
                12, "RNN", 10, 10, n_epochs=20, work_dir=self.temp_work_dir
            )

            # simulate the case that user interrupted training with Ctrl-C after 10 epochs
            model1.fit(self.series, epochs=10)
            self.assertEqual(10, model1.epochs_trained)

            model1.fit(self.series)
            self.assertEqual(20, model1.epochs_trained)

        # n_epochs = 20, fit|epochs=15, epochs_trained=10 - train for 15 epochs
        def test_train_from_10_n_epochs_20_fit_15_epochs(self):
            model1 = RNNModel(
                12, "RNN", 10, 10, n_epochs=20, work_dir=self.temp_work_dir
            )

            # simulate the case that user interrupted training with Ctrl-C after 10 epochs
            model1.fit(self.series, epochs=10)
            self.assertEqual(10, model1.epochs_trained)

            model1.fit(self.series, epochs=15)
            self.assertEqual(15, model1.epochs_trained)

        def test_load_weights_from_checkpoint(self):
            ts_training = self.series[:90]
            ts_test = self.series[90:]
            original_model_name = "original"
            retrained_model_name = "retrained"
            # original model, checkpoints are saved
            model = RNNModel(
                12,
                "RNN",
                5,
                1,
                n_epochs=5,
                work_dir=self.temp_work_dir,
                save_checkpoints=True,
                model_name=original_model_name,
                random_state=1,
            )
            model.fit(ts_training)
            original_preds = model.predict(10)
            original_mape = mape(original_preds, ts_test)

            # load last checkpoint of original model, train it for 2 additional epochs
            model_rt = RNNModel(
                12,
                "RNN",
                5,
                1,
                n_epochs=5,
                work_dir=self.temp_work_dir,
                model_name=retrained_model_name,
                random_state=1,
            )
            model_rt.load_weights_from_checkpoint(
                model_name=original_model_name, work_dir=self.temp_work_dir, best=False
            )

            # must indicate series otherwise self.training_series must be saved in checkpoint
            loaded_preds = model_rt.predict(10, ts_training)
            # save/load checkpoint should produce identical predictions
            self.assertEqual(original_preds, loaded_preds)

            model_rt.fit(ts_training)
            retrained_preds = model_rt.predict(10)
            retrained_mape = mape(retrained_preds, ts_test)
            self.assertTrue(
                retrained_mape < original_mape,
                f"Retrained model has a greater error (mape) than the original model, "
                f"respectively {retrained_mape} and {original_mape}",
            )

            # raise Exception when trying to load ckpt weights in different architecture
            with self.assertRaises(ValueError):
                model_rt = RNNModel(
                    12,
                    "RNN",
                    10,  # loaded model has only 5 hidden_layers
                    5,
                )
                model_rt.load_weights_from_checkpoint(
                    model_name=original_model_name,
                    work_dir=self.temp_work_dir,
                    best=False,
                )

            # raise Exception when trying to pass `weights_only`=True to `torch.load()`
            with self.assertRaises(ValueError):
                model_rt = RNNModel(
                    12,
                    "RNN",
                    5,
                    5,
                )
                model_rt.load_weights_from_checkpoint(
                    model_name=original_model_name,
                    work_dir=self.temp_work_dir,
                    best=False,
                    weights_only=True,
                )

        def test_load_weights(self):
            ts_training = self.series[:90]
            ts_test = self.series[90:]
            original_model_name = "original"
            retrained_model_name = "retrained"
            # original model, checkpoints are saved
            model = RNNModel(
                12,
                "RNN",
                5,
                1,
                n_epochs=5,
                work_dir=self.temp_work_dir,
                save_checkpoints=False,
                model_name=original_model_name,
                random_state=1,
            )
            model.fit(ts_training)
            path_manual_save = os.path.join(self.temp_work_dir, "RNN_manual_save.pt")
            model.save(path_manual_save)
            original_preds = model.predict(10)
            original_mape = mape(original_preds, ts_test)

            # load last checkpoint of original model, train it for 2 additional epochs
            model_rt = RNNModel(
                12,
                "RNN",
                5,
                1,
                n_epochs=5,
                work_dir=self.temp_work_dir,
                model_name=retrained_model_name,
                random_state=1,
            )
            model_rt.load_weights(path=path_manual_save)

            # must indicate series otherwise self.training_series must be saved in checkpoint
            loaded_preds = model_rt.predict(10, ts_training)
            # save/load checkpoint should produce identical predictions
            self.assertEqual(original_preds, loaded_preds)

            model_rt.fit(ts_training)
            retrained_preds = model_rt.predict(10)
            retrained_mape = mape(retrained_preds, ts_test)
            self.assertTrue(
                retrained_mape < original_mape,
                f"Retrained model has a greater mape error than the original model, "
                f"respectively {retrained_mape} and {original_mape}",
            )

        def test_optimizers(self):

            optimizers = [
                (torch.optim.Adam, {"lr": 0.001}),
                (torch.optim.SGD, {"lr": 0.001}),
            ]

            for optim_cls, optim_kwargs in optimizers:
                model = RNNModel(
                    12,
                    "RNN",
                    10,
                    10,
                    optimizer_cls=optim_cls,
                    optimizer_kwargs=optim_kwargs,
                )
                # should not raise an error
                model.fit(self.series, epochs=1)

        def test_lr_schedulers(self):

            lr_schedulers = [
                (torch.optim.lr_scheduler.StepLR, {"step_size": 10}),
                (
                    torch.optim.lr_scheduler.ReduceLROnPlateau,
                    {"threshold": 0.001, "monitor": "train_loss"},
                ),
                (torch.optim.lr_scheduler.ExponentialLR, {"gamma": 0.09}),
            ]

            for lr_scheduler_cls, lr_scheduler_kwargs in lr_schedulers:
                model = RNNModel(
                    12,
                    "RNN",
                    10,
                    10,
                    lr_scheduler_cls=lr_scheduler_cls,
                    lr_scheduler_kwargs=lr_scheduler_kwargs,
                )
                # should not raise an error
                model.fit(self.series, epochs=1)

        def test_wrong_model_creation_params(self):
            valid_kwarg = {"pl_trainer_kwargs": {}}
            invalid_kwarg = {"some_invalid_kwarg": None}

            # valid params should not raise an error
            _ = RNNModel(12, "RNN", 10, 10, **valid_kwarg)

            # invalid params should raise an error
            with self.assertRaises(ValueError):
                _ = RNNModel(12, "RNN", 10, 10, **invalid_kwarg)

        def test_metrics(self):
            metric = MeanAbsolutePercentageError()
            metric_collection = MetricCollection(
                [MeanAbsolutePercentageError(), MeanAbsoluteError()]
            )

            # test single metric
            model = RNNModel(12, "RNN", 10, 10, n_epochs=1, torch_metrics=metric)
            model.fit(self.series)

            # test metric collection
            model = RNNModel(
                12, "RNN", 10, 10, n_epochs=1, torch_metrics=metric_collection
            )
            model.fit(self.series)

            # test multivariate series
            model = RNNModel(12, "RNN", 10, 10, n_epochs=1, torch_metrics=metric)
            model.fit(self.multivariate_series)

        def test_metrics_w_likelihood(self):
            metric = MeanAbsolutePercentageError()
            metric_collection = MetricCollection(
                [MeanAbsolutePercentageError(), MeanAbsoluteError()]
            )

            # test single metric
            model = RNNModel(
                12,
                "RNN",
                10,
                10,
                n_epochs=1,
                likelihood=GaussianLikelihood(),
                torch_metrics=metric,
            )
            model.fit(self.series)

            # test metric collection
            model = RNNModel(
                12,
                "RNN",
                10,
                10,
                n_epochs=1,
                likelihood=GaussianLikelihood(),
                torch_metrics=metric_collection,
            )
            model.fit(self.series)

            # test multivariate series
            model = RNNModel(
                12,
                "RNN",
                10,
                10,
                n_epochs=1,
                likelihood=GaussianLikelihood(),
                torch_metrics=metric_collection,
            )
            model.fit(self.multivariate_series)

        def test_invalid_metrics(self):
            torch_metrics = ["invalid"]
            with self.assertRaises(AttributeError):
                model = RNNModel(
                    12, "RNN", 10, 10, n_epochs=1, torch_metrics=torch_metrics
                )
                model.fit(self.series)

        @pytest.mark.slow
        def test_lr_find(self):
            train_series, val_series = self.series[:-40], self.series[-40:]
            model = RNNModel(12, "RNN", 10, 10, random_state=42)
            # find the learning rate
            res = model.lr_find(series=train_series, val_series=val_series, epochs=50)
            assert isinstance(res, _LRFinder)
            assert res.suggestion() is not None
            # verify that learning rate finder bypasses the `fit` logic
            assert model.model is None
            assert not model._fit_called
            # cannot predict with an untrained model
            with pytest.raises(ValueError):
                model.predict(n=3, series=self.series)

            # check that results are reproducible
            model = RNNModel(12, "RNN", 10, 10, random_state=42)
            res2 = model.lr_find(series=train_series, val_series=val_series, epochs=50)
            assert res.suggestion() == res2.suggestion()

            # check that suggested learning rate is better than the worst
            lr_worst = res.results["lr"][np.argmax(res.results["loss"])]
            lr_suggested = res.suggestion()
            scores = {}
            for lr, lr_name in zip([lr_worst, lr_suggested], ["worst", "suggested"]):
                model = RNNModel(
                    12,
                    "RNN",
                    10,
                    10,
                    n_epochs=10,
                    random_state=42,
                    optimizer_cls=torch.optim.Adam,
                    optimizer_kwargs={"lr": lr},
                )
                model.fit(train_series)
                scores[lr_name] = mape(
                    val_series, model.predict(len(val_series), series=train_series)
                )
            assert scores["worst"] > scores["suggested"]



"""
Difference Scorer
-----------------

This scorer simply computes the elementwise difference
between two series. If the two series are multivariate, it
returns a multivariate series.
"""

from darts.ad.scorers.scorers import NonFittableAnomalyScorer
from darts.timeseries import TimeSeries


class DifferenceScorer(NonFittableAnomalyScorer):
    def __init__(self) -> None:
        super().__init__(univariate_scorer=False, window=1)

    def __str__(self):
        return "Difference"

    def _score_core_from_prediction(
        self,
        actual_series: TimeSeries,
        pred_series: TimeSeries,
    ) -> TimeSeries:
        self._assert_deterministic(actual_series, "actual_series")
        self._assert_deterministic(pred_series, "pred_series")
        return actual_series - pred_series



import shutil
import tempfile
from itertools import product

import numpy as np
import pandas as pd
import pytest

from darts import concatenate
from darts.logging import get_logger
from darts.metrics import rmse
from darts.tests.base_test_class import DartsBaseTestClass
from darts.utils import timeseries_generation as tg

logger = get_logger(__name__)

try:
    import torch

    from darts.models.forecasting.dlinear import DLinearModel
    from darts.models.forecasting.nlinear import NLinearModel
    from darts.utils.likelihood_models import GaussianLikelihood

    TORCH_AVAILABLE = True

except ImportError:
    logger.warning("Torch not available. Dlinear and NLinear tests will be skipped.")
    TORCH_AVAILABLE = False

if TORCH_AVAILABLE:

    class DlinearNlinearModelsTestCase(DartsBaseTestClass):
        np.random.seed(42)
        torch.manual_seed(42)

        def setUp(self):
            self.temp_work_dir = tempfile.mkdtemp(prefix="darts")

        def tearDown(self):
            shutil.rmtree(self.temp_work_dir)

        def test_creation(self):
            with self.assertRaises(ValueError):
                DLinearModel(
                    input_chunk_length=1,
                    output_chunk_length=1,
                    normalize=True,
                    likelihood=GaussianLikelihood(),
                )

            with self.assertRaises(ValueError):
                NLinearModel(
                    input_chunk_length=1,
                    output_chunk_length=1,
                    normalize=True,
                    likelihood=GaussianLikelihood(),
                )

        def test_fit(self):
            large_ts = tg.constant_timeseries(length=100, value=1000)
            small_ts = tg.constant_timeseries(length=100, value=10)

            for (model_cls, kwargs) in [
                (DLinearModel, {"kernel_size": 5}),
                (DLinearModel, {"kernel_size": 6}),
                (NLinearModel, {}),
            ]:
                # Test basic fit and predict
                model = model_cls(
                    input_chunk_length=1,
                    output_chunk_length=1,
                    n_epochs=10,
                    random_state=42,
                    **kwargs
                )
                model.fit(large_ts[:98])
                pred = model.predict(n=2).values()[0]

                # Test whether model trained on one series is better than one trained on another
                model2 = model_cls(
                    input_chunk_length=1,
                    output_chunk_length=1,
                    n_epochs=10,
                    random_state=42,
                )
                model2.fit(small_ts[:98])
                pred2 = model2.predict(n=2).values()[0]
                self.assertTrue(abs(pred2 - 10) < abs(pred - 10))

                # test short predict
                pred3 = model2.predict(n=1)
                self.assertEqual(len(pred3), 1)

        def test_logtensorboard(self):
            ts = tg.constant_timeseries(length=50, value=10)

            for model_cls in [DLinearModel, NLinearModel]:
                # Test basic fit and predict
                model = model_cls(
                    input_chunk_length=1,
                    output_chunk_length=1,
                    n_epochs=1,
                    log_tensorboard=True,
                    work_dir=self.temp_work_dir,
                    pl_trainer_kwargs={"log_every_n_steps": 1},
                )
                model.fit(ts)
                model.predict(n=2)

        def test_shared_weights(self):
            ts = tg.constant_timeseries(length=50, value=10).stack(
                tg.gaussian_timeseries(length=50)
            )

            for model_cls in [DLinearModel, NLinearModel]:
                # Test basic fit and predict
                model_shared = model_cls(
                    input_chunk_length=5,
                    output_chunk_length=1,
                    n_epochs=2,
                    const_init=False,
                    shared_weights=True,
                    random_state=42,
                )
                model_not_shared = model_cls(
                    input_chunk_length=5,
                    output_chunk_length=1,
                    n_epochs=2,
                    const_init=False,
                    shared_weights=False,
                    random_state=42,
                )
                model_shared.fit(ts)
                model_not_shared.fit(ts)
                pred_shared = model_shared.predict(n=2)
                pred_not_shared = model_not_shared.predict(n=2)
                self.assertTrue(
                    np.any(np.not_equal(pred_shared.values(), pred_not_shared.values()))
                )

        def test_multivariate_and_covariates(self):
            np.random.seed(42)
            torch.manual_seed(42)
            # test on multiple multivariate series with future and static covariates

            def _create_multiv_series(f1, f2, n1, n2, nf1, nf2):
                bases = [
                    tg.sine_timeseries(
                        length=400, value_frequency=f, value_amplitude=1.0
                    )
                    for f in (f1, f2)
                ]
                noises = [tg.gaussian_timeseries(length=400, std=n) for n in (n1, n2)]
                noise_modulators = [
                    tg.sine_timeseries(length=400, value_frequency=nf)
                    + tg.constant_timeseries(length=400, value=1) / 2
                    for nf in (nf1, nf2)
                ]
                noises = [noises[i] * noise_modulators[i] for i in range(len(noises))]

                target = concatenate(
                    [bases[i] + noises[i] for i in range(len(bases))], axis="component"
                )

                target = target.with_static_covariates(
                    pd.DataFrame([[f1, n1, nf1], [f2, n2, nf2]])
                )

                return target, concatenate(noise_modulators, axis="component")

            def _eval_model(
                train1,
                train2,
                val1,
                val2,
                fut_cov1,
                fut_cov2,
                cls=DLinearModel,
                lkl=None,
            ):
                model = cls(
                    input_chunk_length=50,
                    output_chunk_length=10,
                    shared_weights=False,
                    const_init=True,
                    likelihood=lkl,
                    random_state=42,
                )

                model.fit(
                    [train1, train2],
                    future_covariates=[fut_cov1, fut_cov2]
                    if fut_cov1 is not None
                    else None,
                    epochs=10,
                )

                pred1, pred2 = model.predict(
                    series=[train1, train2],
                    future_covariates=[fut_cov1, fut_cov2]
                    if fut_cov1 is not None
                    else None,
                    n=len(val1),
                    num_samples=500 if lkl is not None else 1,
                )

                return rmse(val1, pred1), rmse(val2, pred2)

            series1, fut_cov1 = _create_multiv_series(0.05, 0.07, 0.2, 0.4, 0.02, 0.03)
            series2, fut_cov2 = _create_multiv_series(0.04, 0.03, 0.4, 0.1, 0.02, 0.04)

            train1, val1 = series1.split_after(0.7)
            train2, val2 = series2.split_after(0.7)

            for model, lkl in product(
                [DLinearModel, NLinearModel], [None, GaussianLikelihood()]
            ):

                e1, e2 = _eval_model(
                    train1, train2, val1, val2, fut_cov1, fut_cov2, cls=model, lkl=lkl
                )
                self.assertLessEqual(e1, 0.34)
                self.assertLessEqual(e2, 0.28)

                e1, e2 = _eval_model(
                    train1.with_static_covariates(None),
                    train2.with_static_covariates(None),
                    val1,
                    val2,
                    fut_cov1,
                    fut_cov2,
                    cls=model,
                    lkl=lkl,
                )
                self.assertLessEqual(e1, 0.32)
                self.assertLessEqual(e2, 0.28)

                e1, e2 = _eval_model(
                    train1, train2, val1, val2, None, None, cls=model, lkl=lkl
                )
                self.assertLessEqual(e1, 0.40)
                self.assertLessEqual(e2, 0.34)

                e1, e2 = _eval_model(
                    train1.with_static_covariates(None),
                    train2.with_static_covariates(None),
                    val1,
                    val2,
                    None,
                    None,
                    cls=model,
                    lkl=lkl,
                )
                self.assertLessEqual(e1, 0.40)
                self.assertLessEqual(e2, 0.34)

        def test_optional_static_covariates(self):
            series = tg.sine_timeseries(length=20).with_static_covariates(
                pd.DataFrame({"a": [1]})
            )
            for model_cls in [NLinearModel, DLinearModel]:
                # training model with static covs and predicting without will raise an error
                model = model_cls(
                    input_chunk_length=12,
                    output_chunk_length=6,
                    use_static_covariates=True,
                    n_epochs=1,
                )
                model.fit(series)
                with pytest.raises(ValueError):
                    model.predict(n=2, series=series.with_static_covariates(None))

                # with `use_static_covariates=False`, static covariates are ignored and prediction works
                model = model_cls(
                    input_chunk_length=12,
                    output_chunk_length=6,
                    use_static_covariates=False,
                    n_epochs=1,
                )
                model.fit(series)
                preds = model.predict(n=2, series=series.with_static_covariates(None))
                assert preds.static_covariates is None

                # with `use_static_covariates=False`, static covariates are ignored and prediction works
                model = model_cls(
                    input_chunk_length=12,
                    output_chunk_length=6,
                    use_static_covariates=False,
                    n_epochs=1,
                )
                model.fit(series.with_static_covariates(None))
                preds = model.predict(n=2, series=series)
                assert preds.static_covariates.equals(series.static_covariates)



"""
Mapper and InvertibleMapper
---------------------------
"""

from typing import Any, Callable, Mapping, Union

import numpy as np
import pandas as pd

from darts.logging import get_logger
from darts.timeseries import TimeSeries

from .base_data_transformer import BaseDataTransformer
from .invertible_data_transformer import InvertibleDataTransformer

logger = get_logger(__name__)

MapperFn = Union[
    Callable[[np.number], np.number], Callable[[pd.Timestamp, np.number], np.number]
]


class Mapper(BaseDataTransformer):
    def __init__(
        self,
        fn: Union[
            Callable[[np.number], np.number],
            Callable[[pd.Timestamp, np.number], np.number],
        ],
        name: str = "Mapper",
        n_jobs: int = 1,
        verbose: bool = False,
    ):
        """
        Data transformer to apply a custom function to a (sequence of) ``TimeSeries``
        (similar to calling :func:`TimeSeries.map()` on each series).

        The mapper takes care of parallelizing the operations on multiple series over
        multiple processors.

        Parameters
        ----------
        fn
            Either a function which takes a value and returns a value ie. `f(x) = y`
            Or a function which takes a value and its timestamp and returns a value ie. `f(timestamp, x) = y`.
        name
            A specific name for the transformer.
        n_jobs
            The number of jobs to run in parallel. Parallel jobs are created only when a ``Sequence[TimeSeries]`` is
            passed as input to a method, parallelising operations regarding different ``TimeSeries``. Defaults to `1`
            (sequential). Setting the parameter to `-1` means using all the available processors.
            Note: for a small amount of data, the parallelisation overhead could end up increasing the total
            required amount of time.
        verbose
            Optionally, whether to print operations progress

        Examples
        --------
        >>> import numpy as np
        >>> from darts import TimeSeries
        >>> from darts.dataprocessing.transformers import InvertibleMapper
        >>> series = TimeSeries.from_values(np.array([1e0, 1e1, 1e2, 1e3]))
        >>> transformer = InvertibleMapper(np.log10, lambda x: 10**x)
        >>> series_transformed = transformer.transform(series)
        >>> print(series_transformed)
        <TimeSeries (DataArray) (time: 4, component: 1, sample: 1)>
        array([[[0.]],
            [[1.]],
            [[2.]],
            [[3.]]])
        Coordinates:
        * time       (time) int64 0 1 2 3
        * component  (component) <U1 '0'
        Dimensions without coordinates: sample
        """
        # Define fixed params (i.e. attributes defined before calling `super().__init__`):
        self._fn = fn
        super().__init__(name=name, n_jobs=n_jobs, verbose=verbose)

    @staticmethod
    def ts_transform(series: TimeSeries, params: Mapping[str, Any]) -> TimeSeries:
        return series.map(params["fixed"]["_fn"])


class InvertibleMapper(InvertibleDataTransformer):
    def __init__(
        self,
        fn: Union[
            Callable[[np.number], np.number],
            Callable[[pd.Timestamp, np.number], np.number],
        ],
        inverse_fn: Union[
            Callable[[np.number], np.number],
            Callable[[pd.Timestamp, np.number], np.number],
        ],
        name: str = "InvertibleMapper",
        n_jobs: int = 1,
        verbose: bool = False,
    ):
        """
        Data transformer to apply a custom function and its inverse to a (sequence of) ``TimeSeries``
        (similar to calling :func:`TimeSeries.map()` on each series).

        Parameters
        ----------
        fn
            Either a function which takes a value and returns a value ie. `f(x) = y`
            Or a function which takes a value and its timestamp and returns a value ie. `f(timestamp, x) = y`.
        inverse_fn
            Similarly to `fn`, either a function which takes a value and returns a value ie. `f(x) = y`
            Or a function which takes a value and its timestamp and returns a value ie. `f(timestamp, x) = y`.
            `inverse_fn` should be such that ``inverse_fn(fn(x)) == x``.
        name
            A specific name for the transformer.
        n_jobs
            The number of jobs to run in parallel. Parallel jobs are created only when a `Sequence[TimeSeries]` is
            passed as input to a method, parallelising operations regarding different `TimeSeries`. Defaults to `1`
            (sequential). Setting the parameter to `-1` means using all the available processors.
            Note: for a small amount of data, the parallelisation overhead could end up increasing the total
            required amount of time.
        verbose
            Optionally, whether to print operations progress

        Examples
        --------
        >>> import numpy as np
        >>> from darts import TimeSeries
        >>> from darts.dataprocessing.transformers import Mapper
        >>> series = TimeSeries.from_values(np.array([1e0, 1e1, 1e2, 1e3]))
        >>> transformer = Mapper(np.log10)
        >>> series_transformed = transformer.transform(series)
        >>> print(series_transformed)
        <TimeSeries (DataArray) (time: 4, component: 1, sample: 1)>
        array([[[0.]],
            [[1.]],
            [[2.]],
            [[3.]]])
        Coordinates:
        * time       (time) int64 0 1 2 3
        * component  (component) <U1 '0'
        Dimensions without coordinates: sample
        >>> series_restaured = transformer.inverse_transform(series_transformed)
        >>> print(series_restaured)
        <TimeSeries (DataArray) (time: 4, component: 1, sample: 1)>
        array([[[   1.]],
            [[  10.]],
            [[ 100.]],
            [[1000.]]])
        Coordinates:
        * time       (time) int64 0 1 2 3
        * component  (component) <U1 '0'
        Dimensions without coordinates: sample
        """

        self._fn = fn
        self._inverse_fn = inverse_fn
        super().__init__(name=name, n_jobs=n_jobs, verbose=verbose)

    @staticmethod
    def ts_transform(
        series: TimeSeries, params: Mapping[str, Mapping[str, MapperFn]]
    ) -> TimeSeries:
        return series.map(params["fixed"]["_fn"])

    @staticmethod
    def ts_inverse_transform(
        series: TimeSeries, params: Mapping[str, Mapping[str, MapperFn]]
    ) -> TimeSeries:
        return series.map(params["fixed"]["_inverse_fn"])



"""
StatsForecast utils
-----------
"""

import numpy as np

# In a normal distribution, 68.27 percentage of values lie within one standard deviation of the mean
one_sigma_rule = 68.27


def create_normal_samples(
    mu: float,
    std: float,
    num_samples: int,
    n: int,
) -> np.array:
    """Generate samples assuming a Normal distribution."""
    samples = np.random.normal(loc=mu, scale=std, size=(num_samples, n)).T
    samples = np.expand_dims(samples, axis=1)
    return samples


def unpack_sf_dict(
    forecast_dict: dict,
):
    """Unpack the dictionary that is returned by the StatsForecast 'predict()' method."""
    mu = forecast_dict["mean"]
    std = forecast_dict[f"hi-{one_sigma_rule}"] - mu
    return mu, std



import warnings
from itertools import product
from typing import Sequence

import pandas as pd

from darts import TimeSeries
from darts.logging import get_logger, raise_log
from darts.tests.base_test_class import DartsBaseTestClass
from darts.utils.data.tabularization import _get_feature_times
from darts.utils.timeseries_generation import linear_timeseries


class GetFeatureTimesTestCase(DartsBaseTestClass):
    """
    Tests `_get_feature_times` function defined in `darts.utils.data.tabularization`. There
    are broadly two 'groups' of tests defined in this module:
        1. 'Generated Test Cases': these test that `_get_feature_times` produces the same outputs
        as a simplified implementation of this same function. For these tests, the 'correct answer' is not
        directly specified; instead, it is generated from a set of input parameters using the set of simplified
        functions. The rationale behind this approach is that it allows for many different combinations of input
        values to be effortlessly tested. The drawback of this, however, is that the correctness of these tests
        assumes that the simplified functions have been implemented correctly - if this isn't the case, then these
        tests are not to be trusted. In saying this, these simplified functions are significantly easier to
        understand and debug than the `create_lagged_prediction_data` function they're helping to test.
        2. 'Specified Test Cases': these test that `_get_feature_times` returns an exactly specified output; these
        specified outputs are *not* 'generated' by another function. Although these 'specified' test cases tend to
        be simpler and less extensive than the 'generated' test cases, their correctness does not assume the correct
        implementation of any other function.
    """

    #
    #   Helper Functions for Generated Test Cases
    #

    @staticmethod
    def get_feature_times_target_training(
        target_series: TimeSeries,
        lags: Sequence[int],
        output_chunk_length: int,
    ):
        """
        Helper function that returns all the times within `target_series` that can be used to
        create features and labels for training.

        More specifically:
            - The first `max_lag = -min(lags)` times are excluded, since these times
            have fewer than `max_lag` values after them, which means that we can't create
            features for these times.
            - The last `output_chunk_length - 1` times are excluded, since these times don't
            have `(output_chunk_length - 1)` values ahead of them and, therefore, we can't
            create labels for these times.
        """
        times = target_series.time_index
        # Exclude first `max_lag` times:
        max_lag = -min(lags)
        times = times[max_lag:]
        # Exclude last `output_chunk_length - 1` times:
        if output_chunk_length > 1:
            times = times[: -output_chunk_length + 1]
        return times

    @staticmethod
    def get_feature_times_past(
        past_covariates: TimeSeries,
        past_covariates_lags: Sequence[int],
    ) -> pd.Index:
        """
        Helper function that returns all the times within `past_covariates` that can be used to
        create features for training or prediction.

        Unlike the `target_series` during training, features can be constructed for times that
        occur after the end of `past_covariates`; this is because:
            1. We don't need to have all the `past_covariates` values up to time `t` to construct
            a feature for this time; instead, we only need to have the values from time `t - min_lag`
            to `t - max_lag`, where `min_lag = -max(past_covariates_lags)` and
            `max_lag = -min(past_covariates_lags)`. In other words, the latest feature we can create
            for `past_covariates` occurs at `past_covariates.end_time() + min_lag * past_covariates.freq`.
            2. We don't need to use the values of `past_covariates` to construct labels, so we're able
            to create a feature for time `t` without having to worry about whether we can construct
            a corresponding label for this time.
        """
        times = past_covariates.time_index
        min_lag = -max(past_covariates_lags)
        # Add times after end of series for which we can create features:
        times = times.union(
            [times[-1] + i * past_covariates.freq for i in range(1, min_lag + 1)]
        )
        max_lag = -min(past_covariates_lags)
        times = times[max_lag:]
        return times

    @staticmethod
    def get_feature_times_target_prediction(
        target_series: TimeSeries, lags: Sequence[int]
    ):
        """
        Helper function that returns all the times within `target_series` that can be used to
        create features for prediction.

        Since we don't need to worry about creating labels for prediction data, the process
        of constructing prediction features using the `target_series` is identical to
        constructing features for the `past_covariates` series.
        """
        return GetFeatureTimesTestCase.get_feature_times_past(target_series, lags)

    @staticmethod
    def get_feature_times_future(
        future_covariates: TimeSeries,
        future_covariates_lags: Sequence[int],
    ) -> pd.Index:
        """
        Helper function called by `_get_feature_times` that extracts all of the times within
        `future_covariates` that can be used to create features for training or prediction.

        Unlike the lag values for `target_series` and `past_covariates`, the values in
        `future_covariates_lags` can be negative, zero, or positive. This means that
        `min_lag = -max(future_covariates_lags)` and `max_lag = -min(future_covariates_lags)`
        are *not* guaranteed to be positive here: they could be negative (corresponding to
        a positive value in `future_covariates_lags`), zero, or positive (corresponding to
        a negative value in `future_covariates_lags`). With that being said, the relationship
        `min_lag <= max_lag` always holds.

        Consequently, we need to consider three scenarios when finding feature times
        for `future_covariates`:
            1. Both `min_lag` and `max_lag` are positive, which indicates that all of
            the lag values in `future_covariates_lags` are negative (i.e. only values before
            time `t` are used to create a feature from time `t`). In this case, `min_lag`
            and `max_lag` correspond to the smallest magnitude and largest magnitude *negative*
            lags in `future_covariates_lags` respectively. This means we *can* create features for
            times that extend beyond the end of `future_covariates`; additionally, we're unable
            to create features for the first `min_lag` times (see docstring for `get_feature_times_past`).
            2. Both `min_lag` and `max_lag` are non-positive. In this case, `abs(min_lag)` and `abs(max_lag)`
            correspond to the largest and smallest magnitude lags in `future_covariates_lags` respectively;
            note that, somewhat confusingly, `abs(max_lag) <= abs(min_lag)` here. This means that we *can* create f
            features for times that occur before the start of `future_covariates`; the reasoning for this is
            basically the inverse of Case 1 (i.e. we only need to know the values from times `t + abs(max_lag)`
            to `t + abs(min_lag)` to create a feature for time `t`). Additionally, we're unable to create features
            for the last `abs(min_lag)` times in the series, since these times do not have `abs(min_lag)` values
            after them.
            3. `min_lag` is non-positive (i.e. zero or negative), but `max_lag` is positive. In this case,
            `abs(min_lag)` is the magnitude of the largest *non-negative* lag value in `future_covariates_lags`
            and `max_lag` is the largest *negative* lag value in `future_covariates_lags`. This means that we
            *cannot* create features for times that occur before the start of `future_covariates`, nor for
            times that occur after the end of `future_covariates`; this is because we must have access to
            both times before *and* after time `t` to create a feature for this time, which clearly can't
            be acieved for times extending before the start or after the end of the series. Moreover,
            we must exclude the first `max_lag` times and the last `abs(min_lag)` times, since these
            times do not have enough values before or after them respectively.
        """
        times = future_covariates.time_index
        min_lag = -max(future_covariates_lags)
        max_lag = -min(future_covariates_lags)
        # Case 1:
        if (min_lag > 0) and (max_lag > 0):
            # Can create features for times extending after the end of `future_covariates`:
            times = times.union(
                [times[-1] + i * future_covariates.freq for i in range(1, min_lag + 1)]
            )
            # Can't create features for first `max_lag` times in series:
            times = times[max_lag:]
        # Case 2:
        elif (min_lag <= 0) and (max_lag <= 0):
            # Can create features for times before the start of `future_covariates`:
            times = times.union(
                [
                    times[0] - i * future_covariates.freq
                    for i in range(1, abs(max_lag) + 1)
                ]
            )
            # Can't create features for last `abs(min_lag)` times in series:
            times = times[:min_lag] if min_lag != 0 else times
        # Case 3:
        elif (min_lag <= 0) and (max_lag > 0):
            # Can't create features for last `abs(min_lag)` times in series:
            times = times[:min_lag] if min_lag != 0 else times
            # Can't create features for first `max_lag` times in series:
            times = times[max_lag:]
        # Unexpected case:
        else:
            error_msg = (
                "Unexpected `future_covariates_lags` case encountered: "
                "`min_lag` is positive, but `max_lag` is negative. "
                f"Caused by `future_covariates_lags = {future_covariates_lags}`."
            )
            error = ValueError(error_msg)
            raise_log(error, get_logger(__name__))
        return times

    #
    #   Generated Test Cases
    #

    # Input parameter combinations used to generate test cases:
    target_lag_combos = lags_past_combos = (
        [-1],
        [-2, -1],
        [-6, -4, -3],
        [-4, -6, -3],
    )
    lags_future_combos = (*target_lag_combos, [0], [0, 1], [1, 3], [-2, 2])
    ocl_combos = (1, 2, 5, 10)

    def test_feature_times_training_range_idx(self):
        """
        Tests that `_get_feature_times` produces the same `times` output as
        that generated by using the various `get_feature_times_*` helper
        functions defined in this module when `is_training = True`. Consistency
        is checked over all of the combinations of parameter values specified by
        `self.target_lag_combos`, `self.lags_past_combos`, `self.lags_future_combos`
        and `self.max_samples_per_ts_combos`. This particular test uses timeseries
        with range time indices.
        """
        # Define timeseries with different starting points, lengths, and frequencies:
        target = linear_timeseries(start=1, length=20, freq=1)
        past = linear_timeseries(start=2, length=25, freq=2)
        future = linear_timeseries(start=3, length=30, freq=3)
        for (lags, lags_past, lags_future, ocl) in product(
            self.target_lag_combos,
            self.lags_past_combos,
            self.lags_future_combos,
            self.ocl_combos,
        ):
            feature_times = _get_feature_times(
                target_series=target,
                past_covariates=past,
                future_covariates=future,
                lags=lags,
                lags_past_covariates=lags_past,
                lags_future_covariates=lags_future,
                output_chunk_length=ocl,
                is_training=True,
            )
            target_expected = self.get_feature_times_target_training(target, lags, ocl)
            past_expected = self.get_feature_times_past(past, lags_past)
            future_expected = self.get_feature_times_future(future, lags_future)
            self.assertTrue(target_expected.equals(feature_times[0]))
            self.assertTrue(past_expected.equals(feature_times[1]))
            self.assertTrue(future_expected.equals(feature_times[2]))

    def test_feature_times_training_datetime_idx(self):
        """
        Tests that `_get_feature_times` produces the same `times` output as
        that generated by using the various `get_feature_times_*` helper
        functions defined in this module when `is_training = True`. Consistency
        is checked over all of the combinations of parameter values specified by
        `self.target_lag_combos`, `self.lags_past_combos`, `self.lags_future_combos`
        and `self.max_samples_per_ts_combos`. This particular test uses timeseries
        with datetime time indices.
        """
        # Define timeseries with different starting points, lengths, and frequencies:
        target = linear_timeseries(start=pd.Timestamp("1/1/2000"), length=20, freq="1d")
        past = linear_timeseries(start=pd.Timestamp("1/2/2000"), length=25, freq="2d")
        future = linear_timeseries(start=pd.Timestamp("1/3/2000"), length=30, freq="3d")
        for (lags, lags_past, lags_future, ocl) in product(
            self.target_lag_combos,
            self.lags_past_combos,
            self.lags_future_combos,
            self.ocl_combos,
        ):
            feature_times = _get_feature_times(
                target_series=target,
                past_covariates=past,
                future_covariates=future,
                lags=lags,
                lags_past_covariates=lags_past,
                lags_future_covariates=lags_future,
                output_chunk_length=ocl,
                is_training=True,
            )
            target_expected = self.get_feature_times_target_training(target, lags, ocl)
            past_expected = self.get_feature_times_past(past, lags_past)
            future_expected = self.get_feature_times_future(future, lags_future)
            self.assertTrue(target_expected.equals(feature_times[0]))
            self.assertTrue(past_expected.equals(feature_times[1]))
            self.assertTrue(future_expected.equals(feature_times[2]))

    def test_feature_times_prediction_range_idx(self):
        """
        Tests that `_get_feature_times` produces the same `times` output as
        that generated by using the various `get_feature_times_*` helper
        functions defined in this module when `is_training = False` (i.e. when creaiting
        prediction data). Consistency is checked over all of the combinations of parameter
        values specified by `self.target_lag_combos`, `self.lags_past_combos`,
        `self.lags_future_combos` and `self.max_samples_per_ts_combos`. This particular test
        uses timeseries with range time indices.
        """
        # Define timeseries with different starting points, lengths, and frequencies:
        target = linear_timeseries(start=1, length=20, freq=1)
        past = linear_timeseries(start=2, length=25, freq=2)
        future = linear_timeseries(start=3, length=30, freq=3)
        for (lags, lags_past, lags_future) in product(
            self.target_lag_combos, self.lags_past_combos, self.lags_future_combos
        ):
            feature_times = _get_feature_times(
                target_series=target,
                past_covariates=past,
                future_covariates=future,
                lags=lags,
                lags_past_covariates=lags_past,
                lags_future_covariates=lags_future,
                is_training=False,
            )
            target_expected = self.get_feature_times_target_prediction(target, lags)
            past_expected = self.get_feature_times_past(past, lags_past)
            future_expected = self.get_feature_times_future(future, lags_future)
            self.assertTrue(target_expected.equals(feature_times[0]))
            self.assertTrue(past_expected.equals(feature_times[1]))
            self.assertTrue(future_expected.equals(feature_times[2]))

    def test_feature_times_prediction_datetime_idx(self):
        """
        Tests that `_get_feature_times` produces the same `times` output as
        that generated by using the various `get_feature_times_*` helper
        functions defined in this module when `is_training = False` (i.e. when creaiting
        prediction data). Consistency is checked over all of the combinations of parameter
        values specified by `self.target_lag_combos`, `self.lags_past_combos`,
        `self.lags_future_combos` and `self.max_samples_per_ts_combos`. This particular test
        uses timeseries with datetime time indices.
        """
        # Define timeseries with different starting points, lengths, and frequencies:
        target = linear_timeseries(start=pd.Timestamp("1/1/2000"), length=20, freq="1d")
        past = linear_timeseries(start=pd.Timestamp("1/2/2000"), length=25, freq="2d")
        future = linear_timeseries(start=pd.Timestamp("1/3/2000"), length=30, freq="3d")
        for (lags, lags_past, lags_future) in product(
            self.target_lag_combos, self.lags_past_combos, self.lags_future_combos
        ):
            feature_times = _get_feature_times(
                target_series=target,
                past_covariates=past,
                future_covariates=future,
                lags=lags,
                lags_past_covariates=lags_past,
                lags_future_covariates=lags_future,
                is_training=False,
            )
            target_expected = self.get_feature_times_target_prediction(target, lags)
            past_expected = self.get_feature_times_past(past, lags_past)
            future_expected = self.get_feature_times_future(future, lags_future)
            self.assertTrue(target_expected.equals(feature_times[0]))
            self.assertTrue(past_expected.equals(feature_times[1]))
            self.assertTrue(future_expected.equals(feature_times[2]))

    #
    #   Specified Test Cases
    #

    def test_feature_times_output_chunk_length_range_idx(self):
        """
        Tests that the last feature time for the `target_series`
        returned by `_get_feature_times` corresponds to
        `output_chunk_length - 1` timesteps *before* the end of
        the target series; this is the last time point in
        `target_series` which has enough values in front of it
        to create a label. This particular test uses range time
        index series to check this behaviour.
        """
        target = linear_timeseries(start=0, length=20, freq=2)
        # Test multiple `output_chunk_length` values:
        for ocl in (1, 2, 3, 4, 5):
            feature_times = _get_feature_times(
                target_series=target,
                lags=[-2, -3, -5],
                output_chunk_length=ocl,
                is_training=True,
            )
            self.assertEqual(
                feature_times[0][-1], target.end_time() - target.freq * (ocl - 1)
            )

    def test_feature_times_output_chunk_length_datetime_idx(self):
        """
        Tests that the last feature time for the `target_series`
        returned by `_get_feature_times` when `is_training = True`
        corresponds to the time that is `(output_chunk_length - 1)`
        timesteps *before* the end of the target series; this is the
        last time point in `target_series` which has enough values
        in front of it to create a label. This particular test uses
        datetime time index series to check this behaviour.
        """
        target = linear_timeseries(start=pd.Timestamp("1/1/2000"), length=20, freq="2d")
        # Test multiple `output_chunk_length` values:
        for ocl in (1, 2, 3, 4, 5):
            # `is_training = True`
            feature_times = _get_feature_times(
                target_series=target,
                lags=[-2, -3, -5],
                output_chunk_length=ocl,
                is_training=True,
            )
            self.assertEqual(
                feature_times[0][-1], target.end_time() - target.freq * (ocl - 1)
            )

    def test_feature_times_lags_range_idx(self):
        """
        Tests that the first feature time for the `target_series`
        returned by `_get_feature_times` corresponds to the time
        that is `max_lags` timesteps *after* the start of
        the target series; this is the first time point in
        `target_series` which has enough values in preceeding it
        to create a feature. This particular test uses range time
        index series to check this behaviour.
        """
        target = linear_timeseries(start=0, length=20, freq=2)
        # Expect same behaviour when training and predicting:
        for is_training in (False, True):
            for max_lags in (-1, -2, -3, -4, -5):
                feature_times = _get_feature_times(
                    target_series=target,
                    lags=[-1, max_lags],
                    is_training=is_training,
                )
                self.assertEqual(
                    feature_times[0][0],
                    target.start_time() + target.freq * abs(max_lags),
                )

    def test_feature_times_lags_datetime_idx(self):
        """
        Tests that the first feature time for the `target_series`
        returned by `_get_feature_times` corresponds to the time
        that is `max_lags` timesteps *after* the start of
        the target series; this is the first time point in
        `target_series` which has enough values in preceeding it
        to create a feature. This particular test uses datetime time
        index series to check this behaviour.
        """
        target = linear_timeseries(start=pd.Timestamp("1/1/2000"), length=20, freq="2d")
        # Expect same behaviour when training and predicting:
        for is_training in (False, True):
            for max_lags in (-1, -2, -3, -4, -5):
                feature_times = _get_feature_times(
                    target_series=target,
                    lags=[-1, max_lags],
                    is_training=is_training,
                )
                self.assertEqual(
                    feature_times[0][0],
                    target.start_time() + target.freq * abs(max_lags),
                )

    def test_feature_times_training_single_time_range_idx(self):
        """
        Tests that `_get_feature_times` correctly handles case where only
        a single time can be used to create training features and labels.
        This particular test uses range index timeseries.
        """
        # Can only create feature and label for time `1` (`-1` lag behind is time `0`):
        target = linear_timeseries(start=0, length=2, freq=1)
        lags = [-1]
        feature_times = _get_feature_times(
            target_series=target,
            output_chunk_length=1,
            lags=lags,
            is_training=True,
        )
        self.assertEqual(len(feature_times[0]), 1)
        self.assertEqual(feature_times[0][0], 1)

        # Can only create feature for time `6` (`-2` lags behind is time `2`):
        future = linear_timeseries(start=2, length=1, freq=2)
        future_lags = [-2]
        feature_times = _get_feature_times(
            target_series=target,
            future_covariates=future,
            output_chunk_length=1,
            lags=lags,
            lags_future_covariates=future_lags,
            is_training=True,
        )
        self.assertEqual(len(feature_times[0]), 1)
        self.assertEqual(feature_times[0][0], 1)
        self.assertEqual(len(feature_times[2]), 1)
        self.assertEqual(feature_times[2][0], 6)

    def test_feature_times_training_single_time_datetime_idx(self):
        """
        Tests that `_get_feature_times` correctly handles case where only
        a single time can be used to create training features and labels.
        This particular test uses datetime index timeseries.
        """
        # Can only create feature and label for "1/2/2000" (`-1` lag behind is "1/1/2000"):
        target = linear_timeseries(start=pd.Timestamp("1/1/2000"), length=2, freq="d")
        lags = [-1]
        feature_times = _get_feature_times(
            target_series=target,
            output_chunk_length=1,
            lags=lags,
            is_training=True,
        )
        self.assertEqual(len(feature_times[0]), 1)
        self.assertEqual(feature_times[0][0], pd.Timestamp("1/2/2000"))

        # Can only create feature for "1/6/2000" (`-2` lags behind is "1/2/2000"):
        future = linear_timeseries(start=pd.Timestamp("1/2/2000"), length=1, freq="2d")
        future_lags = [-2]
        feature_times = _get_feature_times(
            target_series=target,
            future_covariates=future,
            output_chunk_length=1,
            lags=lags,
            lags_future_covariates=future_lags,
            is_training=True,
        )
        self.assertEqual(len(feature_times[0]), 1)
        self.assertEqual(feature_times[0][0], pd.Timestamp("1/2/2000"))
        self.assertEqual(len(feature_times[2]), 1)
        self.assertEqual(feature_times[2][0], pd.Timestamp("1/6/2000"))

    def test_feature_times_prediction_single_time_range_idx(self):
        """
        Tests that `_get_feature_times` correctly handles case where only
        a single time can be used to create prediction features.
        This particular test uses range index timeseries.
        """
        # Can only create feature for time `1` (`-1` lag behind is time `0`):
        target = linear_timeseries(start=0, length=1, freq=1)
        lags = [-1]
        feature_times = _get_feature_times(
            target_series=target,
            lags=lags,
            is_training=False,
        )
        self.assertEqual(len(feature_times[0]), 1)
        self.assertEqual(feature_times[0][0], 1)

        # Can only create feature for time `6` (`-2` lags behind is time `2`):
        future = linear_timeseries(start=2, length=1, freq=2)
        lags_future = [-2]
        feature_times = _get_feature_times(
            target_series=target,
            future_covariates=future,
            lags=lags,
            lags_future_covariates=lags_future,
            is_training=False,
        )
        self.assertEqual(len(feature_times[0]), 1)
        self.assertEqual(feature_times[0][0], 1)
        self.assertEqual(len(feature_times[2]), 1)
        self.assertEqual(feature_times[2][0], 6)

    def test_feature_times_prediction_single_time_datetime_idx(self):
        """
        Tests that `_get_feature_times` correctly handles case where only
        a single time can be used to create prediction features.
        This particular test uses datetime index timeseries.
        """
        # Can only create feature for "1/2/2000" (`-1` lag behind is time "1/1/2000"):
        target = linear_timeseries(start=pd.Timestamp("1/1/2000"), length=1, freq="d")
        lags = [-1]
        feature_times = _get_feature_times(
            target_series=target,
            lags=lags,
            is_training=False,
        )
        self.assertEqual(len(feature_times[0]), 1)
        self.assertEqual(feature_times[0][0], pd.Timestamp("1/2/2000"))

        # Can only create feature for "1/6/2000" (`-2` lag behind is time "1/2/2000"):
        future = linear_timeseries(start=pd.Timestamp("1/2/2000"), length=1, freq="2d")
        lags_future = [-2]
        feature_times = _get_feature_times(
            target_series=target,
            future_covariates=future,
            lags=lags,
            lags_future_covariates=lags_future,
            is_training=False,
        )
        self.assertEqual(len(feature_times[0]), 1)
        self.assertEqual(feature_times[0][0], pd.Timestamp("1/2/2000"))
        self.assertEqual(len(feature_times[2]), 1)
        self.assertEqual(feature_times[2][0], pd.Timestamp("1/6/2000"))

    def test_feature_times_extend_time_index_range_idx(self):
        """
        Tests that `_get_feature_times` is able to return feature
        times that occur after the end of a series or occur before
        the beginning of a series. This particular test uses range
        index time series.
        """
        # Feature times occur after end of series:
        target = linear_timeseries(start=10, length=1, freq=3)
        past = linear_timeseries(start=2, length=1, freq=2)
        future = linear_timeseries(start=3, length=1, freq=1)
        lags = lags_past = lags_future_1 = [-4]
        feature_times = _get_feature_times(
            target_series=target,
            past_covariates=past,
            future_covariates=future,
            lags=lags,
            lags_past_covariates=lags_past,
            lags_future_covariates=lags_future_1,
            is_training=False,
        )
        self.assertEqual(len(feature_times[0]), 1)
        self.assertEqual(
            feature_times[0][0], target.start_time() - lags[0] * target.freq
        )
        self.assertEqual(len(feature_times[1]), 1)
        self.assertEqual(
            feature_times[1][0], past.start_time() - lags_past[0] * past.freq
        )
        self.assertEqual(len(feature_times[2]), 1)
        self.assertEqual(
            feature_times[2][0], future.start_time() - lags_future_1[0] * future.freq
        )
        # Feature time occurs before start of series:
        lags_future_2 = [4]
        feature_times = _get_feature_times(
            future_covariates=future,
            lags_future_covariates=lags_future_2,
            is_training=False,
        )
        self.assertEqual(len(feature_times[2]), 1)
        self.assertEqual(
            feature_times[2][0], future.start_time() - lags_future_2[0] * future.freq
        )

    def test_feature_times_extend_time_index_datetime_idx(self):
        """
        Tests that `_get_feature_times` is able to return feature
        times that occur after the end of a series or occur before
        the beginning of a series. This particular test uses datetime
        index time series.
        """
        # Feature times occur after end of series:
        target = linear_timeseries(start=pd.Timestamp("1/10/2000"), length=1, freq="3d")
        past = linear_timeseries(start=pd.Timestamp("1/2/2000"), length=1, freq="2d")
        future = linear_timeseries(start=pd.Timestamp("1/3/2000"), length=1, freq="1d")
        lags = lags_past = lags_future_1 = [-4]
        feature_times = _get_feature_times(
            target_series=target,
            past_covariates=past,
            future_covariates=future,
            lags=lags,
            lags_past_covariates=lags_past,
            lags_future_covariates=lags_future_1,
            is_training=False,
        )
        self.assertEqual(len(feature_times[0]), 1)
        self.assertEqual(
            feature_times[0][0], target.start_time() - lags[0] * target.freq
        )
        self.assertEqual(len(feature_times[1]), 1)
        self.assertEqual(
            feature_times[1][0], past.start_time() - lags_past[0] * past.freq
        )
        self.assertEqual(len(feature_times[2]), 1)
        self.assertEqual(
            feature_times[2][0], future.start_time() - lags_future_1[0] * future.freq
        )
        # Feature time occurs before start of series:
        lags_future_2 = [4]
        feature_times = _get_feature_times(
            future_covariates=future,
            lags_future_covariates=lags_future_2,
            is_training=False,
        )
        self.assertEqual(len(feature_times[2]), 1)
        self.assertEqual(
            feature_times[2][0], future.start_time() - lags_future_2[0] * future.freq
        )

    def test_feature_times_future_lags_range_idx(self):
        """
        Tests that `_get_feature_times` correctly handles the `lags_future_covariates`
        argument for the following three cases:
            1. `lags_future_covariates` contains only `0`
            2. `lags_future_covariates` contains only a positive lag
            3. `lags_future_covariates` contains a combination of positive,
            zero, and negative lags
        This particular test uses range index timeseries.
        """
        future = linear_timeseries(start=0, length=10, freq=2)
        # Case 1 - Zero lag:
        lags_future = [0]
        feature_times = _get_feature_times(
            future_covariates=future,
            lags_future_covariates=lags_future,
            is_training=False,
        )
        # All times will be feature times:
        self.assertEqual(len(feature_times[2]), future.n_timesteps)
        self.assertTrue(feature_times[2].equals(future.time_index))

        # Case 2 - Positive lag:
        lags_future = [1]
        feature_times = _get_feature_times(
            future_covariates=future,
            lags_future_covariates=lags_future,
            is_training=False,
        )
        # Need to include new time at start of series; only last time will be excluded:
        extended_future = future.prepend_values([0])
        self.assertEqual(len(feature_times[2]), extended_future.n_timesteps - 1)
        self.assertTrue(feature_times[2].equals(extended_future.time_index[:-1]))

        # Case 3 - Combo of negative, zero, and positive lags:
        lags_future = [-1, 0, 1]
        feature_times = _get_feature_times(
            future_covariates=future,
            lags_future_covariates=lags_future,
            is_training=False,
        )
        # Only first and last times will be excluded:
        self.assertEqual(len(feature_times[2]), future.n_timesteps - 2)
        self.assertTrue(feature_times[2].equals(future.time_index[1:-1]))

    def test_feature_times_future_lags_datetime_idx(self):
        """
        Tests that `_get_feature_times` correctly handles the `lags_future_covariates`
        argument for the following three cases:
            1. `lags_future_covariates` contains only `0`
            2. `lags_future_covariates` contains only a positive lag
            3. `lags_future_covariates` contains a combination of positive,
            zero, and negative lags
        This particular test uses datetime index timeseries.
        """
        future = linear_timeseries(start=pd.Timestamp("1/1/2000"), length=10, freq="2d")
        # Case 1 - Zero lag:
        lags_future = [0]
        feature_times = _get_feature_times(
            future_covariates=future,
            lags_future_covariates=lags_future,
            is_training=False,
        )
        # All times will be feature times:
        self.assertEqual(len(feature_times[2]), future.n_timesteps)
        self.assertTrue(feature_times[2].equals(future.time_index))

        # Case 2 - Positive lag:
        lags_future = [1]
        feature_times = _get_feature_times(
            future_covariates=future,
            lags_future_covariates=lags_future,
            is_training=False,
        )
        # Need to include new time at start of series; only last time will be excluded:
        extended_future = future.prepend_values([0])
        self.assertEqual(len(feature_times[2]), extended_future.n_timesteps - 1)
        self.assertTrue(feature_times[2].equals(extended_future.time_index[:-1]))

        # Case 3 - Combo of negative, zero, and positive lags:
        lags_future = [-1, 0, 1]
        feature_times = _get_feature_times(
            future_covariates=future,
            lags_future_covariates=lags_future,
            is_training=False,
        )
        # Only first and last times will be excluded:
        self.assertEqual(len(feature_times[2]), future.n_timesteps - 2)
        self.assertTrue(feature_times[2].equals(future.time_index[1:-1]))

    def test_feature_times_unspecified_series(self):
        """
        Tests that `_get_feature_times` correctly returns
        `None` in place of a sequence of times if a particular
        series is not specified.
        """
        # Generate simple examples:
        target = linear_timeseries(start=1, length=20, freq=1)
        past = linear_timeseries(start=2, length=25, freq=2)
        future = linear_timeseries(start=3, length=30, freq=3)
        lags = [-1]
        lags_past = [-2]
        lags_future = [-3]
        # Need to extend series then exclude first few starting times:
        expected_target = target.append_values([0]).time_index[1:]
        expected_past = past.append_values(2 * [0]).time_index[2:]
        expected_future = future.append_values(3 * [0]).time_index[3:]

        # Specify only target, without past and future:
        feature_times = _get_feature_times(
            target_series=target, lags=lags, is_training=False
        )
        self.assertTrue(expected_target.equals(feature_times[0]))
        self.assertEqual(feature_times[1], None)
        self.assertEqual(feature_times[2], None)

        # Specify only past, without target and future:
        feature_times = _get_feature_times(
            past_covariates=past,
            lags_past_covariates=lags_past,
            is_training=False,
        )
        self.assertEqual(feature_times[0], None)
        self.assertTrue(expected_past.equals(feature_times[1]))
        self.assertEqual(feature_times[2], None)

        # Specify only future, without target and past:
        feature_times = _get_feature_times(
            future_covariates=future,
            lags_future_covariates=lags_future,
            is_training=False,
        )
        self.assertEqual(feature_times[0], None)
        self.assertEqual(feature_times[1], None)
        self.assertTrue(expected_future.equals(feature_times[2]))

        # Specify target and past, without future:
        feature_times = _get_feature_times(
            target_series=target,
            past_covariates=past,
            lags=lags,
            lags_past_covariates=lags_past,
            is_training=False,
        )
        self.assertTrue(expected_target.equals(feature_times[0]))
        self.assertTrue(expected_past.equals(feature_times[1]))
        self.assertEqual(feature_times[2], None)

        # Specify target and future, without past:
        feature_times = _get_feature_times(
            target_series=target,
            future_covariates=future,
            lags=lags,
            lags_future_covariates=lags_future,
            is_training=False,
        )
        self.assertTrue(expected_target.equals(feature_times[0]))
        self.assertEqual(feature_times[1], None)
        self.assertTrue(expected_future.equals(feature_times[2]))

        # Specify past and future, without target:
        feature_times = _get_feature_times(
            past_covariates=past,
            future_covariates=future,
            lags_past_covariates=lags_past,
            lags_future_covariates=lags_future,
            is_training=False,
        )
        self.assertEqual(feature_times[0], None)
        self.assertTrue(expected_past.equals(feature_times[1]))
        self.assertTrue(expected_future.equals(feature_times[2]))

    def test_feature_times_unspecified_lag_or_series_warning(self):
        """
        Tests that `_get_feature_times` throws correct warning when
        a series is specified by its corresponding lag is not, or
        vice versa. The only circumstance under which a warning
        should *not* be issued is when `target_series` is specified,
        but `lags` is not when `is_training = True`; this is because
        the user may not want to add auto-regressive features to `X`,
        but they still need to specify `target_series` to create labels.
        """
        # Define some arbitrary input values:
        target = linear_timeseries(start=1, length=20, freq=1)
        past = linear_timeseries(start=2, length=25, freq=2)
        future = linear_timeseries(start=3, length=30, freq=3)
        lags = [-1, -2]
        lags_past = [-2, -5]
        lags_future = [-3, -5]
        # Specify `future_covariates` but not `lags_future_covariates` when `is_training = False`
        with warnings.catch_warnings(record=True) as w:
            _ = _get_feature_times(
                past_covariates=past,
                future_covariates=future,
                lags_past_covariates=lags_past,
                is_training=False,
            )
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[0].category, UserWarning))
            self.assertEqual(
                str(w[0].message),
                (
                    "`future_covariates` was specified without accompanying "
                    "`lags_future_covariates` and, thus, will be ignored."
                ),
            )
        # Specify `lags_future_covariates` but not `future_covariates` when `is_training = False`
        with warnings.catch_warnings(record=True) as w:
            _ = _get_feature_times(
                past_covariates=past,
                lags_past_covariates=lags_past,
                lags_future_covariates=lags_future,
                is_training=False,
            )
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[0].category, UserWarning))
            self.assertEqual(
                str(w[0].message),
                (
                    "`lags_future_covariates` was specified without accompanying "
                    "`future_covariates` and, thus, will be ignored."
                ),
            )
        # Specify `future_covariates` but not `lags_future_covariates` and
        # `target_series` but not lags, when `is_training = True`
        with warnings.catch_warnings(record=True) as w:
            _ = _get_feature_times(
                target_series=target,
                past_covariates=past,
                future_covariates=future,
                lags_past_covariates=lags_past,
                output_chunk_length=1,
                is_training=True,
            )
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[0].category, UserWarning))
            self.assertEqual(
                str(w[0].message),
                (
                    "`future_covariates` was specified without accompanying "
                    "`lags_future_covariates` and, thus, will be ignored."
                ),
            )
        # Specify `lags_future_covariates` but not `future_covariates`, and
        # `target_series` but not lags, when `is_training = True`
        with warnings.catch_warnings(record=True) as w:
            _ = _get_feature_times(
                target_series=target,
                past_covariates=past,
                lags_past_covariates=lags_past,
                lags_future_covariates=lags_future,
                output_chunk_length=1,
                is_training=True,
            )
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[0].category, UserWarning))
            self.assertEqual(
                str(w[0].message),
                (
                    "`lags_future_covariates` was specified without accompanying "
                    "`future_covariates` and, thus, will be ignored."
                ),
            )
        # Specify `lags_future_covariates` but not `future_covariates`, and
        # `past_covariates` but not `lags_past_covariates`, when `is_training = True`
        with warnings.catch_warnings(record=True) as w:
            _ = _get_feature_times(
                target_series=target,
                past_covariates=past,
                lags=lags,
                lags_future_covariates=lags_future,
                is_training=False,
            )
            self.assertEqual(len(w), 2)
            self.assertTrue(issubclass(w[0].category, UserWarning))
            self.assertTrue(issubclass(w[1].category, UserWarning))
            self.assertEqual(
                str(w[0].message),
                (
                    "`past_covariates` was specified without accompanying "
                    "`lags_past_covariates` and, thus, will be ignored."
                ),
            )
            self.assertEqual(
                str(w[1].message),
                (
                    "`lags_future_covariates` was specified without accompanying "
                    "`future_covariates` and, thus, will be ignored."
                ),
            )
        # Specify `lags_future_covariates` but not `future_covariates`, and
        # `past_covariates` but not `lags_past_covariates`, and `target_series`
        # but not `lags` when `is_training = False`:
        with warnings.catch_warnings(record=True) as w:
            _ = _get_feature_times(
                target_series=target,
                past_covariates=past,
                lags_future_covariates=lags_future,
                output_chunk_length=1,
                is_training=True,
            )
            self.assertEqual(len(w), 2)
            self.assertTrue(issubclass(w[0].category, UserWarning))
            self.assertTrue(issubclass(w[1].category, UserWarning))
            self.assertEqual(
                str(w[0].message),
                (
                    "`past_covariates` was specified without accompanying "
                    "`lags_past_covariates` and, thus, will be ignored."
                ),
            )
            self.assertEqual(
                str(w[1].message),
                (
                    "`lags_future_covariates` was specified without accompanying "
                    "`future_covariates` and, thus, will be ignored."
                ),
            )
        # Specify `target_series` but not `lags` when `is_training = False`:
        with warnings.catch_warnings(record=True) as w:
            _ = _get_feature_times(
                target_series=target,
                past_covariates=past,
                future_covariates=future,
                lags_past_covariates=lags_past,
                lags_future_covariates=lags_future,
                is_training=False,
            )
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[0].category, UserWarning))
        # Specify `target_series` but not `lags` when `is_training = True`;
        # this should *not* throw a warning:
        with warnings.catch_warnings(record=True) as w:
            _ = _get_feature_times(
                target_series=target,
                past_covariates=past,
                future_covariates=future,
                lags_past_covariates=lags_past,
                lags_future_covariates=lags_future,
                is_training=True,
            )
            self.assertEqual(len(w), 0)

    def test_feature_times_unspecified_training_inputs_error(self):
        """
        Tests that `_get_feature_times` throws correct error when
        `target_series` and/or `output_chunk_length` hasn't been
        specified when `is_training = True`.
        """
        output_chunk_length = 1
        # Don't specify `target_series`:
        with self.assertRaises(ValueError) as e:
            _get_feature_times(
                output_chunk_length=output_chunk_length, is_training=True
            )
        self.assertEqual(
            ("Must specify `target_series` when `is_training = True`."),
            str(e.exception),
        )
        # Don't specify neither `target_series` nor `output_chunk_length`
        with self.assertRaises(ValueError) as e:
            _get_feature_times(is_training=True)
        self.assertEqual(
            ("Must specify `target_series` when `is_training = True`."),
            str(e.exception),
        )

    def test_feature_times_no_lags_specified_error(self):
        """
        Tests that `_get_feature_times` throws correct error
        when no lags have been specified.
        """
        target = linear_timeseries(start=1, length=20, freq=1)
        with self.assertRaises(ValueError) as e:
            _get_feature_times(target_series=target, is_training=False)
        self.assertEqual(
            "Must specify at least one of: `lags`, `lags_past_covariates`, `lags_future_covariates`.",
            str(e.exception),
        )

    def test_feature_times_series_too_short_error(self):
        """
        Tests that `_get_feature_times` throws correct error
        when provided series are too short for specified
        lag and/or `output_chunk_length` values.
        """
        series = linear_timeseries(start=1, length=2, freq=1)
        # `target_series` too short when predicting:
        with self.assertRaises(ValueError) as e:
            _get_feature_times(target_series=series, lags=[-20, -1], is_training=False)
        self.assertEqual(
            (
                "`target_series` must have at least `-min(lags) + max(lags) + 1` = 20 "
                "timesteps; instead, it only has 2."
            ),
            str(e.exception),
        )
        # `target_series` too short when training:
        with self.assertRaises(ValueError) as e:
            _get_feature_times(
                target_series=series,
                lags=[-20],
                output_chunk_length=5,
                is_training=True,
            )
        self.assertEqual(
            (
                "`target_series` must have at least `-min(lags) + output_chunk_length` = 25 "
                "timesteps; instead, it only has 2."
            ),
            str(e.exception),
        )
        # `past_covariates` too short when training:
        with self.assertRaises(ValueError) as e:
            _get_feature_times(
                target_series=series,
                past_covariates=series,
                lags_past_covariates=[-20, -1],
                output_chunk_length=1,
                is_training=True,
            )
        self.assertEqual(
            (
                "`past_covariates` must have at least "
                "`-min(lags_past_covariates) + max(lags_past_covariates) + 1` = 20 timesteps; "
                "instead, it only has 2."
            ),
            str(e.exception),
        )

    def test_feature_times_invalid_lag_values_error(self):
        """
        Tests that `_get_feature_times` throws correct error
        when provided with invalid lag values (i.e. not less than
        0 if `lags`, or not less than 1 if `lags_past_covariates` or
        `lags_future_covariates`).
        """
        series = linear_timeseries(start=1, length=3, freq=1)
        # `lags` not <= -1:
        with self.assertRaises(ValueError) as e:
            _get_feature_times(target_series=series, lags=[0], is_training=False)
        self.assertEqual(
            ("`lags` must be a `Sequence` containing only `int` values less than 0."),
            str(e.exception),
        )
        # `lags_past_covariates` not <= -1:
        with self.assertRaises(ValueError) as e:
            _get_feature_times(
                past_covariates=series, lags_past_covariates=[0], is_training=False
            )
        self.assertEqual(
            (
                "`lags_past_covariates` must be a `Sequence` containing only `int` values less than 0."
            ),
            str(e.exception),
        )
        # `lags_future_covariates` can be positive, negative, and/or zero - no error should be thrown:
        _get_feature_times(
            future_covariates=series,
            lags_future_covariates=[-1, 0, 1],
            is_training=False,
        )



"""
Forecasting Model Explainer Base Class

A forecasting model explainer takes a fitted forecasting model as input and applies an Explainability model
to it. Its purpose is to explain each past input contribution to a given model forecast. This 'explanation'
depends on the characteristics of the XAI model chosen (shap, lime etc...).

"""
from abc import ABC, abstractmethod
from typing import Collection, Optional, Sequence, Union

from darts import TimeSeries
from darts.explainability.explainability_result import ExplainabilityResult
from darts.logging import get_logger, raise_if, raise_if_not, raise_log
from darts.models.forecasting.forecasting_model import ForecastingModel
from darts.utils.statistics import stationarity_tests
from darts.utils.utils import series2seq

logger = get_logger(__name__)

MIN_BACKGROUND_SAMPLE = 10


class ForecastingModelExplainer(ABC):
    @abstractmethod
    def __init__(
        self,
        model: ForecastingModel,
        background_series: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        background_past_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
        background_future_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
    ):
        """
        The base class for forecasting model explainers. It defines the *minimal* behavior that all
        forecasting model explainers support.

        Naming:

        - A background series is a `TimeSeries` with which to 'train' the `Explainer` model.
        - A foreground series is the `TimeSeries` to explain using the fitted `Explainer` model.

        Parameters
        ----------
        model
            A `ForecastingModel` to be explained. It must be fitted first.
        background_series
            A series or list of series to *train* the `ForecastingModelExplainer` along with any foreground series.
            Consider using a reduced well-chosen background to reduce computation time.

                - optional if `model` was fit on a single target series. By default, it is the `series` used
                at fitting time.
                - mandatory if `model` was fit on multiple (sequence of) target series.

        background_past_covariates
            A past covariates series or list of series that the model needs once fitted.
        background_future_covariates
            A future covariates series or list of series that the model needs once fitted.
        """
        if not model._fit_called:
            raise_log(
                ValueError(
                    "The model must be fitted before instantiating a ForecastingModelExplainer."
                ),
                logger,
            )

        if model._is_probabilistic():
            logger.warning(
                "The model is probabilistic, but num_samples=1 will be used for explainability."
            )

        self.model = model

        # if `background_series` was not passed, use `training_series` saved in fitted forecasting model.
        if background_series is None:

            raise_if(
                (background_past_covariates is not None)
                or (background_future_covariates is not None),
                "Supplied background past or future covariates but no background series. Please provide "
                "`background_series`.",
            )

            raise_if(
                self.model.training_series is None,
                "`background_series` must be provided if `model` was fit on multiple time series.",
            )

            background_series = self.model.training_series
            background_past_covariates = self.model.past_covariate_series
            background_future_covariates = self.model.future_covariate_series

        else:
            if self.model.encoders.encoding_available:
                (
                    background_past_covariates,
                    background_future_covariates,
                ) = self.model.generate_fit_encodings(
                    series=background_series,
                    past_covariates=background_past_covariates,
                    future_covariates=background_future_covariates,
                )

        self.background_series = series2seq(background_series)
        self.background_past_covariates = series2seq(background_past_covariates)
        self.background_future_covariates = series2seq(background_future_covariates)

        if self.model.supports_past_covariates:
            raise_if(
                self.model.uses_past_covariates
                and self.background_past_covariates is None,
                "A background past covariates is not provided, but the model needs past covariates.",
            )

        if self.model.supports_future_covariates:
            raise_if(
                self.model.uses_future_covariates
                and self.background_future_covariates is None,
                "A background future covariates is not provided, but the model needs future covariates.",
            )

        self.target_components = self.background_series[0].columns.to_list()
        self.past_covariates_components = None
        if self.background_past_covariates is not None:
            self.past_covariates_components = self.background_past_covariates[
                0
            ].columns.to_list()
        self.future_covariates_components = None
        if self.background_future_covariates is not None:
            self.future_covariates_components = self.background_future_covariates[
                0
            ].columns.to_list()

        self._check_background_covariates(
            self.background_series,
            self.background_past_covariates,
            self.background_future_covariates,
            self.target_components,
            self.past_covariates_components,
            self.future_covariates_components,
        )

        if not self._test_stationarity():
            logger.warning(
                "At least one time series component of the background time series is not stationary."
                " Beware of wrong interpretation with chosen explainability."
            )

    @staticmethod
    def _check_background_covariates(
        background_series,
        background_past_covariates,
        background_future_covariates,
        target_components,
        past_covariates_components,
        future_covariates_components,
    ) -> None:

        if background_past_covariates is not None:
            raise_if_not(
                len(background_series) == len(background_past_covariates),
                "The number of background series and past covariates must be the same.",
            )

        if background_future_covariates is not None:
            raise_if_not(
                len(background_series) == len(background_future_covariates),
                "The number of background series and future covariates must be the same.",
            )

        # ensure we have the same names between TimeSeries (if list of). Important to ensure homogeneity
        # for explained features.
        for idx in range(len(background_series)):
            raise_if_not(
                all(
                    [
                        background_series[idx].columns.to_list() == target_components,
                        background_past_covariates[idx].columns.to_list()
                        == past_covariates_components
                        if background_past_covariates is not None
                        else True,
                        background_future_covariates[idx].columns.to_list()
                        == future_covariates_components
                        if background_future_covariates is not None
                        else True,
                    ]
                ),
                "Columns names must be identical between TimeSeries list components (multi-TimeSeries).",
            )

    @abstractmethod
    def explain(
        self,
        foreground_series: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        foreground_past_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
        foreground_future_covariates: Optional[
            Union[TimeSeries, Sequence[TimeSeries]]
        ] = None,
        horizons: Optional[Collection[int]] = None,
        target_components: Optional[Collection[str]] = None,
    ) -> ExplainabilityResult:
        """
        Explains a foreground time series, returns an :class:`ExplainabilityResult`.

        Results can be retrieved via the method
        :func:`ExplainabilityResult.get_explanation(horizon, target_component)`.
        The result is a multivariate `TimeSeries` instance containing the 'explanation'
        for the (horizon, target_component) forecast at any timestamp forecastable corresponding to
        the foreground `TimeSeries` input.

        The component name convention of this multivariate `TimeSeries` is:
        ``"{name}_{type_of_cov}_lag_{idx}"``, where:

        - ``{name}`` is the component name from the original foreground series (target, past, or future).
        - ``{type_of_cov}`` is the covariates type. It can take 3 different values:
          ``"target"``, ``"past_cov"`` or ``"future_cov"``.
        - ``{idx}`` is the lag index.

        **Example:**

        Say we have a model with 2 target components named ``"T_0"`` and ``"T_1"``,
        3 past covariates with default component names ``"0"``, ``"1"``, and ``"2"``,
        and one future covariate with default component name ``"0"``.
        Also, ``horizons = [1, 2]``.
        The model is a regression model, with ``lags = 3``, ``lags_past_covariates=[-1, -3]``,
        ``lags_future_covariates = [0]``.

        We provide `foreground_series`, `foreground_past_covariates`, `foreground_future_covariates` each of length 5.


        >>> explain_results = explainer.explain(
        >>>     foreground_series=foreground_series,
        >>>     foreground_past_covariates=foreground_past_covariates,
        >>>     foreground_future_covariates=foreground_future_covariates,
        >>>     horizons=[1, 2],
        >>>     target_names=["T_0", "T_1"])
        >>> output = explain_results.get_explanation(horizon=1, target="T_1")

        Then the method returns a multivariate TimeSeries containing the *explanations* of
        the corresponding `ForecastingModelExplainer`, with the following component names:

             - T_0_target_lag-1
             - T_0_target_lag-2
             - T_0_target_lag-3
             - T_1_target_lag-1
             - T_1_target_lag-2
             - T_1_target_lag-3
             - 0_past_cov_lag-1
             - 0_past_cov_lag-3
             - 1_past_cov_lag-1
             - 1_past_cov_lag-3
             - 2_past_cov_lag-1
             - 2_past_cov_lag-3
             - 0_fut_cov_lag_0

        This series has length 3, as the model can explain 5-3+1 forecasts
        (timestamp indexes 4, 5, and 6)

        Parameters
        ----------
        foreground_series
            Optionally, the target `TimeSeries` to be explained. Can be multivariate.
            If not provided, the background `TimeSeries` will be explained instead.
        foreground_past_covariates
            Optionally, past covariate timeseries if needed by the ForecastingModel.
        foreground_future_covariates
            Optionally, future covariate timeseries if needed by the ForecastingModel.
        horizons
            Optionally, a collection of integers representing the future lags to be explained.
            Horizon 1 corresponds to the first timestamp being forecasted.
            All values must be no larger than `output_chunk_length` of the explained model.
        target_components
            Optionally, A list of string naming the target components to be explained.

         Returns
         -------
         ExplainabilityResult
             The forecast explanations.

        """
        pass

    def _test_stationarity(self):
        return all(
            [
                (
                    stationarity_tests(background_serie[c])
                    for c in background_serie.components
                )
                for background_serie in self.background_series
            ]
        )



"""
Anomaly Aggregators
-------------------

An anomaly aggregator can take multiple detected anomalies
(in the form of binary TimeSeries, as coming from an anomaly detector)
and combine them into one. It can typically be used to combine
the detections of multiple models into one final detection.

The key method is ``predict()``, which takes as input one (or multiple)
multivariate binary TimeSeries where each component represents the
detection of a single model, and returns one (or multiple) univariate
binary TimeSeries representing the final detection.
"""

from .and_aggregator import AndAggregator
from .ensemble_sklearn_aggregator import EnsembleSklearnAggregator
from .or_aggregator import OrAggregator



"""
Pipeline
--------
"""
from copy import deepcopy
from typing import Iterator, Sequence, Union

from darts import TimeSeries
from darts.dataprocessing.transformers import (
    BaseDataTransformer,
    FittableDataTransformer,
    InvertibleDataTransformer,
)
from darts.logging import get_logger, raise_if_not

logger = get_logger(__name__)


class Pipeline:
    def __init__(
        self,
        transformers: Sequence[BaseDataTransformer],
        copy: bool = False,
        verbose: bool = None,
        n_jobs: int = None,
    ):
        """
        Pipeline to combine multiple data transformers, chaining them together.

        Parameters
        ----------
        transformers
            Sequence of data transformers.
        copy
            If set makes a (deep) copy of each data transformer before adding them to the pipeline
        n_jobs
            The number of jobs to run in parallel. Parallel jobs are created only when a ``Sequence[TimeSeries]`` is
            passed as input to a method, parallelising operations regarding different ``TimeSeries``. Defaults to `1`
            (sequential). Setting the parameter to `-1` means using all the available processors.
            Note: for a small amount of data, the parallelisation overhead could end up increasing the total
            required amount of time.
            Note: this parameter will overwrite the value set in each single transformer. Leave this parameter set to
            `None` for keeping the original transformers' configurations.
        verbose
            Whether to print progress of the operations.
            Note: this parameter will overwrite the value set in each single transformer. Leave this parameter set
            to `None` for keeping the transformers configurations.

        Examples
        --------
        >>> import numpy as np
        >>> from darts import TimeSeries
        >>> from darts.datasets import AirPassengersDataset
        >>> from darts.dataprocessing.transformers import Scaler, MissingValuesFiller
        >>> from darts.dataprocessing.pipeline import Pipeline
        >>> values = np.arange(start=0, stop=12.5, step=2.5)
        >>> values[1:3] = np.nan
        >>> series = series.from_values(values)
        >>> pipeline = Pipeline([MissingValuesFiller(), Scaler()])
        >>> series_transformed = pipeline.fit_transform(series)
        <TimeSeries (DataArray) (time: 5, component: 1, sample: 1)>
        array([[[0.  ]],
            [[0.25]],
            [[0.5 ]],
            [[0.75]],
            [[1.  ]]])
        Coordinates:
        * time       (time) int64 0 1 2 3 4
        * component  (component) object '0'
        Dimensions without coordinates: sample
        """

        raise_if_not(
            all((isinstance(t, BaseDataTransformer)) for t in transformers),
            "transformers should be objects deriving from BaseDataTransformer",
            logger,
        )

        if transformers is None or len(transformers) == 0:
            logger.warning("Empty pipeline created")
            self._transformers: Sequence[BaseDataTransformer[TimeSeries]] = []
        elif copy:
            self._transformers = deepcopy(transformers)
        else:
            self._transformers = transformers

        self._invertible = all(
            isinstance(t, InvertibleDataTransformer) for t in self._transformers
        )

        if verbose is not None:
            for transformer in self._transformers:
                transformer.set_verbose(verbose)

        if n_jobs is not None:
            for transformer in self._transformers:
                transformer.set_n_jobs(n_jobs)

    def fit(self, data: Union[TimeSeries, Sequence[TimeSeries]]):
        """
        Fit all fittable transformers in pipeline.

        Parameters
        ----------
        data
            (`Sequence` of) `TimeSeries` to fit on.
        """

        # Find the last fittable transformer index
        # No need to fit (and thus transform) after this index, possibly saving a fair bit of time
        last_fittable_idx = -1
        for idx, transformer in enumerate(self._transformers):
            if isinstance(transformer, FittableDataTransformer):
                last_fittable_idx = idx

        for idx, transformer in enumerate(self._transformers):
            if idx <= last_fittable_idx and isinstance(
                transformer, FittableDataTransformer
            ):
                transformer.fit(data)

            if idx < last_fittable_idx:
                data = transformer.transform(data)

    def fit_transform(
        self, data: Union[TimeSeries, Sequence[TimeSeries]]
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:
        """
        For each data transformer in the pipeline, first fit the data if transformer is fittable then transform data
        using fitted transformer. The transformed data is then passed to next transformer.

        Parameters
        ----------
        data
            (`Sequence` of) `TimeSeries` to fit and transform on.

        Returns
        -------
        Union[TimeSeries, Sequence[TimeSeries]]
            Transformed data.
        """
        for transformer in self._transformers:
            if isinstance(transformer, FittableDataTransformer):
                transformer.fit(data)

            data = transformer.transform(data)
        return data

    def transform(
        self, data: Union[TimeSeries, Sequence[TimeSeries]]
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:
        """
        For each data transformer in pipeline transform data. Then transformed data is passed to next transformer.

        Parameters
        ----------
        data
            (`Sequence` of) `TimeSeries` to be transformed.

        Returns
        -------
        Union[TimeSeries, Sequence[TimeSeries]]
            Transformed data.
        """
        for transformer in self._transformers:
            data = transformer.transform(data)
        return data

    def inverse_transform(
        self, data: Union[TimeSeries, Sequence[TimeSeries]], partial: bool = False
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:
        """
        For each data transformer in the pipeline, inverse-transform data. Then inverse transformed data is passed to
        the next transformer. Transformers are traversed in reverse order. Raises value error if not all of the
        transformers are invertible and ``partial`` is set to False. Set ``partial`` to True for inverting only the
        InvertibleDataTransformer in the pipeline.

        Parameters
        ----------
        data
            (Sequence of) TimeSeries to be inverse transformed.
        partial
            If set to `True`, the inverse transformation is applied even if the pipeline is not fully invertible,
            calling `inverse_transform()` only on the `InvertibleDataTransformer`s

        Returns
        -------
        Union[TimeSeries, Sequence[TimeSeries]]
            Inverse transformed data.
        """
        if not partial:
            raise_if_not(
                self._invertible,
                "Not all transformers in the pipeline can perform inverse_transform",
                logger,
            )

            for transformer in reversed(self._transformers):
                data = transformer.inverse_transform(data)
            return data
        else:
            for transformer in reversed(self._transformers):
                if isinstance(transformer, InvertibleDataTransformer):
                    data = transformer.inverse_transform(data)
            return data

    def invertible(self) -> bool:
        """
        Returns whether the pipeline is invertible or not.
        A pipeline is invertible if all transformers in the pipeline are themselves invertible.

        Returns
        -------
        bool
            `True` if the pipeline is invertible, `False` otherwise
        """
        return self._invertible

    def __getitem__(self, key: Union[int, slice]) -> "Pipeline":
        """
        Gets subset of Pipeline based either on index or slice with indexes.
        Resulting pipeline will deep copy transformers of the original pipeline.

        Parameters
        ----------
        key
            Either int or slice indicating the subset of data transformers to keep.

        Returns
        -------
        Pipeline
            Subset of pipeline determined by key.
        """
        raise_if_not(
            isinstance(key, int) or isinstance(key, slice),
            "key must be either an int or a slice",
            logger,
        )

        if isinstance(key, int):
            transformers = [self._transformers[key]]
        else:
            transformers = self._transformers[key]
        return Pipeline(transformers, copy=True)

    def __iter__(self) -> Iterator[BaseDataTransformer]:
        """
        Returns
        -------
        Iterator
            Iterator on sequence of data transformers
        """
        return self._transformers.__iter__()

    def __len__(self):
        return len(self._transformers)

    def __copy__(self, deep: bool = True):
        return Pipeline(self._transformers, copy=deep)

    def __deepcopy__(self, memo=None):
        return self.__copy__(deep=True)

    def __str__(self):
        string = "Pipeline: "
        arrow = " -> "
        for transformer in self._transformers:
            string += str(transformer) + arrow

        return string[: -len(arrow)]

    def __repr__(self):
        return self.__str__()



from typing import Sequence

import numpy as np

from darts import TimeSeries
from darts.ad.detectors.quantile_detector import QuantileDetector
from darts.ad.detectors.threshold_detector import ThresholdDetector
from darts.tests.base_test_class import DartsBaseTestClass

list_NonFittableDetector = [ThresholdDetector(low_threshold=0.2)]

list_FittableDetector = [QuantileDetector(low_quantile=0.2)]


class ADDetectorsTestCase(DartsBaseTestClass):

    np.random.seed(42)

    # univariate series
    np_train = np.random.normal(loc=10, scale=0.5, size=100)
    train = TimeSeries.from_values(np_train)

    np_test = np.random.normal(loc=10, scale=1, size=100)
    test = TimeSeries.from_times_and_values(train._time_index, np_test)

    np_anomalies = np.random.choice(a=[0, 1], size=100, p=[0.9, 0.1])
    anomalies = TimeSeries.from_times_and_values(train._time_index, np_anomalies)

    # multivariate series
    np_mts_train = np.random.normal(loc=[10, 5], scale=[0.5, 1], size=[100, 2])
    mts_train = TimeSeries.from_values(np_mts_train)

    np_mts_test = np.random.normal(loc=[10, 5], scale=[1, 1.5], size=[100, 2])
    mts_test = TimeSeries.from_times_and_values(mts_train._time_index, np_mts_test)

    np_mts_anomalies = np.random.choice(a=[0, 1], size=[100, 2], p=[0.9, 0.1])
    mts_anomalies = TimeSeries.from_times_and_values(
        mts_train._time_index, np_mts_anomalies
    )

    np_probabilistic = np.random.choice(a=[0, 1], p=[0.5, 0.5], size=[100, 1, 5])
    probabilistic = TimeSeries.from_values(np_probabilistic)

    def test_DetectNonFittableDetector(self):

        detector = ThresholdDetector(low_threshold=0.2)

        # Check return types
        # Check if return TimeSeries is float when input is a series
        self.assertTrue(isinstance(detector.detect(self.test), TimeSeries))

        # Check if return type is Sequence when input is a Sequence of series
        self.assertTrue(isinstance(detector.detect([self.test]), Sequence))

        # Check if return TimeSeries is Sequence when input is a multivariate series
        self.assertTrue(isinstance(detector.detect(self.mts_test), TimeSeries))

        # Check if return type is Sequence when input is a multivariate series
        self.assertTrue(isinstance(detector.detect([self.mts_test]), Sequence))

        with self.assertRaises(ValueError):
            # Input cannot be probabilistic
            detector.detect(self.probabilistic)

    def test_DetectFittableDetector(self):
        detector = QuantileDetector(low_quantile=0.2)

        # Check return types

        detector.fit(self.train)
        # Check if return type is float when input is a series
        self.assertTrue(isinstance(detector.detect(self.test), TimeSeries))

        # Check if return type is Sequence when input is a sequence of series
        self.assertTrue(isinstance(detector.detect([self.test]), Sequence))

        detector.fit(self.mts_train)
        # Check if return type is Sequence when input is a multivariate series
        self.assertTrue(isinstance(detector.detect(self.mts_test), TimeSeries))

        # Check if return type is Sequence when input is a sequence of multivariate series
        self.assertTrue(isinstance(detector.detect([self.mts_test]), Sequence))

        with self.assertRaises(ValueError):
            # Input cannot be probabilistic
            detector.detect(self.probabilistic)

    def test_eval_accuracy(self):

        detector = ThresholdDetector(low_threshold=0.2)

        # Check return types
        # Check if return type is float when input is a series
        self.assertTrue(
            isinstance(detector.eval_accuracy(self.anomalies, self.test), float)
        )

        # Check if return type is Sequence when input is a Sequence of series
        self.assertTrue(
            isinstance(detector.eval_accuracy(self.anomalies, [self.test]), Sequence)
        )

        # Check if return type is Sequence when input is a multivariate series
        self.assertTrue(
            isinstance(
                detector.eval_accuracy(self.mts_anomalies, self.mts_test), Sequence
            )
        )

        # Check if return type is Sequence when input is a multivariate series
        self.assertTrue(
            isinstance(
                detector.eval_accuracy(self.mts_anomalies, [self.mts_test]), Sequence
            )
        )

        with self.assertRaises(ValueError):
            # Input cannot be probabilistic
            detector.eval_accuracy(self.anomalies, self.probabilistic)

    def test_FittableDetector(self):

        for detector in list_FittableDetector:

            # Need to call fit() before calling detect()
            with self.assertRaises(ValueError):
                detector.detect(self.test)

            # Check if _fit_called is False
            self.assertTrue(not detector._fit_called)

            with self.assertRaises(ValueError):
                # fit on sequence with series that have different width
                detector.fit([self.train, self.mts_train])

            with self.assertRaises(ValueError):
                # Input cannot be probabilistic
                detector.fit(self.probabilistic)

            # case1: fit on UTS
            detector1 = detector
            detector1.fit(self.train)

            # Check if _fit_called is True after being fitted
            self.assertTrue(detector1._fit_called)

            with self.assertRaises(ValueError):
                # series must be same width as series used for training
                detector1.detect(self.mts_test)

            # case2: fit on MTS
            detector2 = detector
            detector2.fit(self.mts_test)

            # Check if _fit_called is True after being fitted
            self.assertTrue(detector2._fit_called)

            with self.assertRaises(ValueError):
                # series must be same width as series used for training
                detector2.detect(self.train)

    def test_QuantileDetector(self):

        # Need to have at least one parameter (low, high) not None
        with self.assertRaises(ValueError):
            QuantileDetector()
        with self.assertRaises(ValueError):
            QuantileDetector(low_quantile=None, high_quantile=None)

        # Parameter low must be float or Sequence of float
        with self.assertRaises(TypeError):
            QuantileDetector(low_quantile="0.5")
        with self.assertRaises(TypeError):
            QuantileDetector(low_quantile=[0.2, "0.1"])

        # Parameter high must be float or Sequence of float
        with self.assertRaises(TypeError):
            QuantileDetector(high_quantile="0.5")
        with self.assertRaises(TypeError):
            QuantileDetector(high_quantile=[0.2, "0.1"])

        # if high and low are both sequences of length>1, they must be of the same size
        with self.assertRaises(ValueError):
            QuantileDetector(low_quantile=[0.2, 0.1], high_quantile=[0.95, 0.8, 0.9])
        with self.assertRaises(ValueError):
            QuantileDetector(low_quantile=[0.2, 0.1, 0.7], high_quantile=[0.95, 0.8])

        # Parameter must be between 0 and 1
        with self.assertRaises(ValueError):
            QuantileDetector(high_quantile=1.1)
        with self.assertRaises(ValueError):
            QuantileDetector(high_quantile=-0.2)
        with self.assertRaises(ValueError):
            QuantileDetector(high_quantile=[-0.1, 0.9])
        with self.assertRaises(ValueError):
            QuantileDetector(low_quantile=1.1)
        with self.assertRaises(ValueError):
            QuantileDetector(low_quantile=-0.2)
        with self.assertRaises(ValueError):
            QuantileDetector(low_quantile=[-0.2, 0.3])

        # Parameter high must be higher or equal than parameter low
        with self.assertRaises(ValueError):
            QuantileDetector(low_quantile=0.7, high_quantile=0.2)
        with self.assertRaises(ValueError):
            QuantileDetector(low_quantile=[0.2, 0.9], high_quantile=[0.95, 0.1])
        with self.assertRaises(ValueError):
            QuantileDetector(low_quantile=0.2, high_quantile=[0.95, 0.1])
        with self.assertRaises(ValueError):
            QuantileDetector(low_quantile=[0.2, 0.9], high_quantile=0.8)

        # Parameter high/low cannot be sequence of only None
        with self.assertRaises(ValueError):
            QuantileDetector(low_quantile=[None, None, None])
        with self.assertRaises(ValueError):
            QuantileDetector(high_quantile=[None, None, None])
        with self.assertRaises(ValueError):
            QuantileDetector(low_quantile=[None], high_quantile=[None, None, None])

        # check that low_threshold and high_threshold are the same and no errors are raised
        detector = QuantileDetector(low_quantile=0.5, high_quantile=0.5)
        detector.fit(self.train)
        self.assertEqual(detector.low_threshold, detector.high_threshold)

        # widths of series used for fitting must match the number of values given for high or/and low,
        # if high and low have a length higher than 1

        detector = QuantileDetector(low_quantile=0.1, high_quantile=[0.8, 0.7])
        with self.assertRaises(ValueError):
            detector.fit(self.train)
        with self.assertRaises(ValueError):
            detector.fit([self.train, self.mts_train])

        detector = QuantileDetector(low_quantile=[0.1, 0.2], high_quantile=[0.8, 0.9])
        with self.assertRaises(ValueError):
            detector.fit(self.train)
        with self.assertRaises(ValueError):
            detector.fit([self.train, self.mts_train])

        detector = QuantileDetector(low_quantile=[0.1, 0.2], high_quantile=0.8)
        with self.assertRaises(ValueError):
            detector.fit(self.train)
        with self.assertRaises(ValueError):
            detector.fit([self.train, self.mts_train])

        detector = QuantileDetector(low_quantile=[0.1, 0.2])
        with self.assertRaises(ValueError):
            detector.fit(self.train)
        with self.assertRaises(ValueError):
            detector.fit([self.train, self.mts_train])

        detector = QuantileDetector(high_quantile=[0.1, 0.2])
        with self.assertRaises(ValueError):
            detector.fit(self.train)
        with self.assertRaises(ValueError):
            detector.fit([self.train, self.mts_train])

        # widths of series used for scoring must match the number of values given for high or/and low,
        # if high and low have a length higher than 1

        detector = QuantileDetector(low_quantile=0.1, high_quantile=[0.8, 0.7])
        detector.fit(self.mts_train)
        with self.assertRaises(ValueError):
            detector.detect(self.train)
        with self.assertRaises(ValueError):
            detector.detect([self.train, self.mts_train])

        detector = QuantileDetector(low_quantile=[0.1, 0.2], high_quantile=[0.8, 0.9])
        detector.fit(self.mts_train)
        with self.assertRaises(ValueError):
            detector.detect(self.train)
        with self.assertRaises(ValueError):
            detector.detect([self.train, self.mts_train])

        detector = QuantileDetector(low_quantile=[0.1, 0.2], high_quantile=0.8)
        detector.fit(self.mts_train)
        with self.assertRaises(ValueError):
            detector.detect(self.train)
        with self.assertRaises(ValueError):
            detector.detect([self.train, self.mts_train])

        detector = QuantileDetector(low_quantile=[0.1, 0.2])
        detector.fit(self.mts_train)
        with self.assertRaises(ValueError):
            detector.detect(self.train)
        with self.assertRaises(ValueError):
            detector.detect([self.train, self.mts_train])

        detector = QuantileDetector(high_quantile=[0.1, 0.2])
        detector.fit(self.mts_train)
        with self.assertRaises(ValueError):
            detector.detect(self.train)
        with self.assertRaises(ValueError):
            detector.detect([self.train, self.mts_train])

        detector = QuantileDetector(low_quantile=0.05, high_quantile=0.95)
        detector.fit(self.train)

        binary_detection = detector.detect(self.test)

        # Return of .detect() must be binary
        self.assertTrue(
            np.array_equal(
                binary_detection.values(copy=False),
                binary_detection.values(copy=False).astype(bool),
            )
        )

        # Return of .detect() must be same len as input
        self.assertTrue(len(binary_detection) == len(self.test))

        # univariate test
        # detector parameter 'abs_low_' must be equal to 9.13658 when trained on the series 'train'
        self.assertAlmostEqual(detector.low_threshold[0], 9.13658, delta=1e-05)

        # detector parameter 'abs_high_' must be equal to 10.74007 when trained on the series 'train'
        self.assertAlmostEqual(detector.high_threshold[0], 10.74007, delta=1e-05)

        # detector must found 10 anomalies in the series 'train'
        self.assertEqual(
            detector.detect(self.train).sum(axis=0).all_values().flatten()[0], 10
        )

        # detector must found 42 anomalies in the series 'test'
        self.assertEqual(binary_detection.sum(axis=0).all_values().flatten()[0], 42)

        # detector must have an accuracy of 0.57 for the series 'test'
        self.assertAlmostEqual(
            detector.eval_accuracy(self.anomalies, self.test, metric="accuracy"),
            0.57,
            delta=1e-05,
        )
        # detector must have an recall of 0.4 for the series 'test'
        self.assertAlmostEqual(
            detector.eval_accuracy(self.anomalies, self.test, metric="recall"),
            0.4,
            delta=1e-05,
        )
        # detector must have an f1 of 0.08510 for the series 'test'
        self.assertAlmostEqual(
            detector.eval_accuracy(self.anomalies, self.test, metric="f1"),
            0.08510,
            delta=1e-05,
        )
        # detector must have an precision of 0.04761 for the series 'test'
        self.assertAlmostEqual(
            detector.eval_accuracy(self.anomalies, self.test, metric="precision"),
            0.04761,
            delta=1e-05,
        )

        # multivariate test
        detector_1param = QuantileDetector(low_quantile=0.05, high_quantile=0.95)
        detector_1param.fit(self.mts_train)
        binary_detection = detector_1param.detect(self.mts_test)

        # if two values are given for low and high, and a series of width 2 is given, then the results must
        # be the same as a detector that was given only one value for low and high.
        # (will duplicate the value for each component)
        detector_2param = QuantileDetector(
            low_quantile=[0.05, 0.05], high_quantile=[0.95, 0.95]
        )
        detector_2param.fit(self.mts_train)
        binary_detection_2param = detector_2param.detect(self.mts_test)
        self.assertEqual(binary_detection, binary_detection_2param)

        # width of output must be equal to 2 (same dimension as input)
        self.assertEqual(binary_detection.width, 2)
        self.assertEqual(
            len(
                detector_1param.eval_accuracy(
                    self.mts_anomalies, self.mts_test, metric="accuracy"
                )
            ),
            2,
        )
        self.assertEqual(
            len(
                detector_1param.eval_accuracy(
                    self.mts_anomalies, self.mts_test, metric="recall"
                )
            ),
            2,
        )
        self.assertEqual(
            len(
                detector_1param.eval_accuracy(
                    self.mts_anomalies, self.mts_test, metric="f1"
                )
            ),
            2,
        )
        self.assertEqual(
            len(
                detector_1param.eval_accuracy(
                    self.mts_anomalies, self.mts_test, metric="precision"
                )
            ),
            2,
        )

        abs_low_ = detector_1param.low_threshold
        abs_high_ = detector_1param.high_threshold

        # detector_1param parameter 'abs_high_' must be equal to 10.83047 when trained
        # on the series 'train' for the 1st component
        self.assertAlmostEqual(abs_high_[0], 10.83047, delta=1e-05)
        # detector_1param parameter 'abs_high_' must be equal to 6.47822 when trained
        # on the series 'train' for the 2nd component
        self.assertAlmostEqual(abs_high_[1], 6.47822, delta=1e-05)

        # detector_1param parameter 'abs_low_' must be equal to 9.20248 when trained
        # on the series 'train' for the 1st component
        self.assertAlmostEqual(abs_low_[0], 9.20248, delta=1e-05)
        # detector_1param parameter 'abs_low_' must be equal to 3.61853 when trained
        # on the series 'train' for the 2nd component
        self.assertAlmostEqual(abs_low_[1], 3.61853, delta=1e-05)

        # detector_1param must found 37 anomalies on the first component of the series 'test'
        self.assertEqual(
            binary_detection["0"].sum(axis=0).all_values().flatten()[0], 37
        )
        # detector_1param must found 38 anomalies on the second component of the series 'test'
        self.assertEqual(
            binary_detection["1"].sum(axis=0).all_values().flatten()[0], 38
        )

        acc = detector_1param.eval_accuracy(
            self.mts_anomalies, self.mts_test, metric="accuracy"
        )
        # detector_1param must have an accuracy of 0.58 on the first component of the series 'mts_test'
        self.assertAlmostEqual(acc[0], 0.58, delta=1e-05)
        # detector_1param must have an accuracy of 0.58 on the second component of the series 'mts_test'
        self.assertAlmostEqual(acc[1], 0.58, delta=1e-05)

        precision = detector_1param.eval_accuracy(
            self.mts_anomalies, self.mts_test, metric="precision"
        )
        # detector_1param must have an precision of 0.08108 on the first component of the series 'mts_test'
        self.assertAlmostEqual(precision[0], 0.08108, delta=1e-05)
        # detector_1param must have an precision of 0.07894 on the second component of the series 'mts_test'
        self.assertAlmostEqual(precision[1], 0.07894, delta=1e-05)

        recall = detector_1param.eval_accuracy(
            self.mts_anomalies, self.mts_test, metric="recall"
        )
        # detector_1param must have an recall of 0.2727 on the first component of the series 'mts_test'
        self.assertAlmostEqual(recall[0], 0.27272, delta=1e-05)
        # detector_1param must have an recall of 0.3 on the second component of the series 'mts_test'
        self.assertAlmostEqual(recall[1], 0.3, delta=1e-05)

        f1 = detector_1param.eval_accuracy(
            self.mts_anomalies, self.mts_test, metric="f1"
        )
        # detector_1param must have an f1 of 0.125 on the first component of the series 'mts_test'
        self.assertAlmostEqual(f1[0], 0.125, delta=1e-05)
        # detector_1param must have an f1 of 0.125 on the second component of the series 'mts_test'
        self.assertAlmostEqual(f1[1], 0.125, delta=1e-05)

        # exemple multivariate with Nones
        detector = QuantileDetector(
            low_quantile=[0.05, None], high_quantile=[None, 0.95]
        )
        detector.fit(self.mts_train)
        binary_detection = detector.detect(self.mts_test)

        # width of output must be equal to 2 (same dimension as input)
        self.assertEqual(binary_detection.width, 2)
        self.assertEqual(
            len(
                detector.eval_accuracy(
                    self.mts_anomalies, self.mts_test, metric="accuracy"
                )
            ),
            2,
        )
        self.assertEqual(
            len(
                detector.eval_accuracy(
                    self.mts_anomalies, self.mts_test, metric="recall"
                )
            ),
            2,
        )
        self.assertEqual(
            len(detector.eval_accuracy(self.mts_anomalies, self.mts_test, metric="f1")),
            2,
        )
        self.assertEqual(
            len(
                detector.eval_accuracy(
                    self.mts_anomalies, self.mts_test, metric="precision"
                )
            ),
            2,
        )

        # TODO: we should improve these tests to introduce some correlation
        # between actual and detected anomalies...

        # detector must found 20 anomalies on the first component of the series 'test'
        # Note: there are 200 values (100 time step x 2 components) so this matches
        # well a detection rate of 10% (bottom 5% on first component and top 5% on second component)
        self.assertEqual(
            binary_detection["0"].sum(axis=0).all_values().flatten()[0], 20
        )
        # detector must found 19 anomalies on the second component of the series 'test'
        self.assertEqual(
            binary_detection["1"].sum(axis=0).all_values().flatten()[0], 19
        )

        acc = detector.eval_accuracy(
            self.mts_anomalies, self.mts_test, metric="accuracy"
        )
        self.assertAlmostEqual(acc[0], 0.69, delta=1e-05)
        self.assertAlmostEqual(acc[1], 0.75, delta=1e-05)

        precision = detector.eval_accuracy(
            self.mts_anomalies, self.mts_test, metric="precision"
        )
        self.assertAlmostEqual(precision[0], 0.0, delta=1e-05)
        self.assertAlmostEqual(precision[1], 0.10526, delta=1e-05)

        recall = detector.eval_accuracy(
            self.mts_anomalies, self.mts_test, metric="recall"
        )
        self.assertAlmostEqual(recall[0], 0.0, delta=1e-05)
        self.assertAlmostEqual(recall[1], 0.2, delta=1e-05)

        f1 = detector.eval_accuracy(self.mts_anomalies, self.mts_test, metric="f1")
        self.assertAlmostEqual(f1[0], 0.0, delta=1e-05)
        self.assertAlmostEqual(f1[1], 0.13793, delta=1e-05)

    def test_ThresholdDetector(self):

        # Parameters
        # Need to have at least one parameter (low, high) not None
        with self.assertRaises(ValueError):
            ThresholdDetector()
        with self.assertRaises(ValueError):
            ThresholdDetector(low_threshold=None, high_threshold=None)

        # if high and low are both sequences of length>1, they must be of the same size
        with self.assertRaises(ValueError):
            ThresholdDetector(low_threshold=[0.2, 0.1], high_threshold=[0.95, 0.8, 0.9])
        with self.assertRaises(ValueError):
            ThresholdDetector(low_threshold=[0.2, 0.1, 0.7], high_threshold=[0.95, 0.8])

        # Parameter high must be higher or equal than parameter low
        with self.assertRaises(ValueError):
            ThresholdDetector(low_threshold=0.7, high_threshold=0.2)
        with self.assertRaises(ValueError):
            ThresholdDetector(low_threshold=[0.2, 0.9], high_threshold=[0.95, 0.1])
        with self.assertRaises(ValueError):
            ThresholdDetector(low_threshold=0.2, high_threshold=[0.95, 0.1])
        with self.assertRaises(ValueError):
            ThresholdDetector(low_threshold=[0.2, 0.9], high_threshold=0.8)
        with self.assertRaises(ValueError):
            ThresholdDetector(low_threshold=[0.2, 0.9, None], high_threshold=0.8)

        # Parameter high/low cannot be sequence of only None
        with self.assertRaises(ValueError):
            ThresholdDetector(low_threshold=[None, None, None])
        with self.assertRaises(ValueError):
            ThresholdDetector(high_threshold=[None, None, None])
        with self.assertRaises(ValueError):
            ThresholdDetector(low_threshold=[None], high_threshold=[None, None, None])

        # widths of series used for scoring must match the number of values given for high or/and low,
        # if high and low have a length higher than 1

        detector = ThresholdDetector(low_threshold=0.1, high_threshold=[0.8, 0.7])
        with self.assertRaises(ValueError):
            detector.detect(self.train)
        with self.assertRaises(ValueError):
            detector.detect([self.train, self.mts_train])

        detector = ThresholdDetector(
            low_threshold=[0.1, 0.2], high_threshold=[0.8, 0.9]
        )
        with self.assertRaises(ValueError):
            detector.detect(self.train)
        with self.assertRaises(ValueError):
            detector.detect([self.train, self.mts_train])

        detector = ThresholdDetector(low_threshold=[0.1, 0.2], high_threshold=0.8)
        with self.assertRaises(ValueError):
            detector.detect(self.train)
        with self.assertRaises(ValueError):
            detector.detect([self.train, self.mts_train])

        detector = ThresholdDetector(low_threshold=[0.1, 0.2])
        with self.assertRaises(ValueError):
            detector.detect(self.train)
        with self.assertRaises(ValueError):
            detector.detect([self.train, self.mts_train])

        detector = ThresholdDetector(high_threshold=[0.1, 0.2])
        with self.assertRaises(ValueError):
            detector.detect(self.train)
        with self.assertRaises(ValueError):
            detector.detect([self.train, self.mts_train])

        detector = ThresholdDetector(low_threshold=9.5, high_threshold=10.5)
        binary_detection = detector.detect(self.test)

        # Return of .detect() must be binary
        self.assertTrue(
            np.array_equal(
                binary_detection.values(copy=False),
                binary_detection.values(copy=False).astype(bool),
            )
        )

        # Return of .detect() must be same len as input
        self.assertTrue(len(binary_detection) == len(self.test))

        # univariate test
        # detector must found 58 anomalies in the series 'test'
        self.assertEqual(binary_detection.sum(axis=0).all_values().flatten()[0], 58)
        # detector must have an accuracy of 0.41 for the series 'test'
        self.assertAlmostEqual(
            detector.eval_accuracy(self.anomalies, self.test, metric="accuracy"),
            0.41,
            delta=1e-05,
        )
        # detector must have an recall of 0.4 for the series 'test'
        self.assertAlmostEqual(
            detector.eval_accuracy(self.anomalies, self.test, metric="recall"),
            0.4,
            delta=1e-05,
        )
        # detector must have an f1 of 0.06349 for the series 'test'
        self.assertAlmostEqual(
            detector.eval_accuracy(self.anomalies, self.test, metric="f1"),
            0.06349,
            delta=1e-05,
        )
        # detector must have an precision of 0.03448 for the series 'test'
        self.assertAlmostEqual(
            detector.eval_accuracy(self.anomalies, self.test, metric="precision"),
            0.03448,
            delta=1e-05,
        )

        # multivariate test
        detector = ThresholdDetector(low_threshold=4.8, high_threshold=10.5)
        binary_detection = detector.detect(self.mts_test)

        # if two values are given for low and high, and a series of width 2 is given,
        # then the results must be the same as a detector that was given only one value
        # for low and high. (will duplicate the value for each width)
        detector_2param = ThresholdDetector(
            low_threshold=[4.8, 4.8], high_threshold=[10.5, 10.5]
        )
        binary_detection_2param = detector_2param.detect(self.mts_test)
        self.assertEqual(binary_detection, binary_detection_2param)

        # width of output must be equal to 2 (same dimension as input)
        self.assertEqual(binary_detection.width, 2)
        self.assertEqual(
            len(
                detector.eval_accuracy(
                    self.mts_anomalies, self.mts_test, metric="accuracy"
                )
            ),
            2,
        )
        self.assertEqual(
            len(
                detector.eval_accuracy(
                    self.mts_anomalies, self.mts_test, metric="recall"
                )
            ),
            2,
        )
        self.assertEqual(
            len(detector.eval_accuracy(self.mts_anomalies, self.mts_test, metric="f1")),
            2,
        )
        self.assertEqual(
            len(
                detector.eval_accuracy(
                    self.mts_anomalies, self.mts_test, metric="precision"
                )
            ),
            2,
        )

        # detector must found 28 anomalies on the first width of the series 'test'
        self.assertEqual(
            binary_detection["0"].sum(axis=0).all_values().flatten()[0], 28
        )
        # detector must found 52 anomalies on the second width of the series 'test'
        self.assertEqual(
            binary_detection["1"].sum(axis=0).all_values().flatten()[0], 52
        )

        acc = detector.eval_accuracy(
            self.mts_anomalies, self.mts_test, metric="accuracy"
        )
        # detector must have an accuracy of 0.71 on the first width of the series 'mts_test'
        self.assertAlmostEqual(acc[0], 0.71, delta=1e-05)
        # detector must have an accuracy of 0.48 on the second width of the series 'mts_test'
        self.assertAlmostEqual(acc[1], 0.48, delta=1e-05)

        precision = detector.eval_accuracy(
            self.mts_anomalies, self.mts_test, metric="precision"
        )
        # detector must have an precision of 0.17857 on the first width of the series 'mts_test'
        self.assertAlmostEqual(precision[0], 0.17857, delta=1e-05)
        # detector must have an precision of 0.09615 on the second width of the series 'mts_test'
        self.assertAlmostEqual(precision[1], 0.09615, delta=1e-05)

        recall = detector.eval_accuracy(
            self.mts_anomalies, self.mts_test, metric="recall"
        )
        # detector must have an recall of 0.45454 on the first width of the series 'mts_test'
        self.assertAlmostEqual(recall[0], 0.45454, delta=1e-05)
        # detector must have an recall of 0.5 on the second width of the series 'mts_test'
        self.assertAlmostEqual(recall[1], 0.5, delta=1e-05)

        f1 = detector.eval_accuracy(self.mts_anomalies, self.mts_test, metric="f1")
        # detector must have an f1 of 0.25641 on the first width of the series 'mts_test'
        self.assertAlmostEqual(f1[0], 0.25641, delta=1e-05)
        # detector must have an f1 of 0.16129 on the second width of the series 'mts_test'
        self.assertAlmostEqual(f1[1], 0.16129, delta=1e-05)

        # exemple multivariate with Nones
        detector = ThresholdDetector(low_threshold=[10, None], high_threshold=[None, 5])
        binary_detection = detector.detect(self.mts_test)

        # width of output must be equal to 2 (same dimension as input)
        self.assertEqual(binary_detection.width, 2)
        self.assertEqual(
            len(
                detector.eval_accuracy(
                    self.mts_anomalies, self.mts_test, metric="accuracy"
                )
            ),
            2,
        )
        self.assertEqual(
            len(
                detector.eval_accuracy(
                    self.mts_anomalies, self.mts_test, metric="recall"
                )
            ),
            2,
        )
        self.assertEqual(
            len(detector.eval_accuracy(self.mts_anomalies, self.mts_test, metric="f1")),
            2,
        )
        self.assertEqual(
            len(
                detector.eval_accuracy(
                    self.mts_anomalies, self.mts_test, metric="precision"
                )
            ),
            2,
        )

        # detector must found 48 anomalies on the first width of the series 'test'
        self.assertEqual(
            binary_detection["0"].sum(axis=0).all_values().flatten()[0], 48
        )
        # detector must found 43 anomalies on the second width of the series 'test'
        self.assertEqual(
            binary_detection["1"].sum(axis=0).all_values().flatten()[0], 43
        )

        acc = detector.eval_accuracy(
            self.mts_anomalies, self.mts_test, metric="accuracy"
        )
        # detector must have an accuracy of 0.51 on the first width of the series 'mts_test'
        self.assertAlmostEqual(acc[0], 0.51, delta=1e-05)
        # detector must have an accuracy of 0.57 on the second width of the series 'mts_test'
        self.assertAlmostEqual(acc[1], 0.57, delta=1e-05)

        precision = detector.eval_accuracy(
            self.mts_anomalies, self.mts_test, metric="precision"
        )
        # detector must have an precision of 0.10416 and  on the first width of the series 'mts_test'
        self.assertAlmostEqual(precision[0], 0.10416, delta=1e-05)
        # detector must have an precision of 0.11627 on the second width of the series 'mts_test'
        self.assertAlmostEqual(precision[1], 0.11627, delta=1e-05)

        recall = detector.eval_accuracy(
            self.mts_anomalies, self.mts_test, metric="recall"
        )
        # detector must have an recall of 0.45454 on the first width of the series 'mts_test'
        self.assertAlmostEqual(recall[0], 0.45454, delta=1e-05)
        # detector must have an recall of 0.5 on the second width of the series 'mts_test'
        self.assertAlmostEqual(recall[1], 0.5, delta=1e-05)

        f1 = detector.eval_accuracy(self.mts_anomalies, self.mts_test, metric="f1")
        # detector must have an f1 of 0.16949 on the first width of the series 'mts_test'
        self.assertAlmostEqual(f1[0], 0.16949, delta=1e-05)
        # detector must have an f1 of 0.18867 on the second width of the series 'mts_test'
        self.assertAlmostEqual(f1[1], 0.18867, delta=1e-05)

    def test_fit_detect(self):

        detector1 = QuantileDetector(low_quantile=0.05, high_quantile=0.95)
        detector1.fit(self.train)
        prediction1 = detector1.detect(self.train)

        detector2 = QuantileDetector(low_quantile=0.05, high_quantile=0.95)
        prediction2 = detector2.fit_detect(self.train)

        self.assertEqual(prediction1, prediction2)
